/***************************************************************************
 *
 * A SeedLink client for Earthworm.  Collects MSEED data records
 * from a SeedLink server and inserts TRACEBUF2 or MSEED messages into
 * an Earthworm ring.
 *
 * Originally written by Chad Trabant, ORFEUS/EC-Project MEREDIAN
 * The liss2ew sources were used as a starting point, thanks Pete Lombard.
 ***************************************************************************/

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* Earthworm includes */
#include "earthworm.h"
#include "kom.h"
#include "trace_buf.h"
#include "transport.h"

#include "libmseed.h"
#include "libslink/libslink.h"

#include "mseedprocess.h"
#include "slink2ew.h"

#ifdef _WINNT
#include <windows.h>
#define setenv(A, B, C) SetEnvironmentVariable ((A), (B))
#endif

/***************************************************************************
 *
 * Global parameters
 *
 ***************************************************************************/
SLCD *slconn;                      /* SeedLink connection parameters       */
short int verbose = 0;             /* Verbosity level                      */
char *statefile   = 0;             /* State file                           */
int stateint      = 1000;          /* Interval (packets recv'd) to save state */
SHM_INFO regionOut;                /* Shared memory region                 */
pid_t myPid;                       /* Process ID                           */
char ringName[MAX_RING_STR];       /* Name of destination ring for data    */
long ringKey;                      /* Key to output shared memory region   */
char myModName[MAX_MOD_STR];       /* Name of module instance              */
long maxrate   = 0;                /* Max rate in bytes/sec, 0 to disable  */
int writemseed = 0;                /* Write MSEED instead of TRACEBUF2     */
unsigned char myModId;             /* ID of this module                    */
unsigned char myInstId;            /* Installation running this module     */
unsigned char myInstIdConf = 0;    /* Installation ID configuration flag   */
unsigned char typeError;           /* Error message type                   */
unsigned char typeHeartbeat;       /* Heartbeat message type               */
unsigned char typeMseed;           /* Waveform message type MSEED          */
unsigned char typeTracebuf2 = 0;   /* Waveform message type TRACEBUF2      */
MSG_LOGO hrtLogo;                  /* Heartbeat message logo               */
MSG_LOGO waveLogo;                 /* Waveform message logo                */
MSG_LOGO errLogo;                  /* Error message logo                   */
int heartbeatInt;                  /* Heartbeat interval (seconds)         */
int logSwitch;                     /* 1 -> write log, 0 -> no log
                                    * 2 -> write module log but not stderr/stdout */
char mseedRingName[MAX_RING_STR];  /* Name of destination ring for mseed data     */
long mseedRingKey;                 /* Key to output shared memory region   */
SHM_INFO *mregionOutP, mregionOut; /* Shared memory region (mseedRing)     */
double MaxLatency      = 0.0;      /* new max latency test, to prevent old data floods */
char auth_buffer[1024] = {0};      /* Buffer for SeedLink authentication   */

int
main (int argc, char **argv)
{
  static char plbuffer[SL_MAX_PAYLOAD];  /* payload buffer */
  const SLpacketinfo *packetinfo = NULL; /* packet information */
  int retval;
  int packetcnt = 0;
  time_t timeNow;          /* current time                  */
  time_t timeLastBeat = 0; /* time last heartbeat was sent  */
  double etimeLastPkt;     /* time last packet was received */
  double etimeThisPkt;     /* time this packet was received */

  int prethrottle  = 0;
  int throttlecnt  = 1;
  int throttlewait = 0;

  /* Allocate and initialize a new connection description */
  slconn = sl_initslcd ("slink2ew", VERSION);

  /* Check number of command line arguments */
  if (argc != 2)
  {
    usage (argv[0]);
    exit (EW_FAILURE);
  }

  /* Process configuration parameters */
  configure (argv);

  /* Attach to Output transport rings */
  tport_attach (&regionOut, ringKey);
  if (mseedRingName[0])
  {
    if (ringKey == mseedRingKey)
      mregionOutP = &regionOut;
    else
    {
      mregionOutP = &mregionOut;
      tport_attach (mregionOutP, mseedRingKey);
    }
  }
  else
    mregionOutP = NULL;

  logit ("t", "slink2ew version %s\n", VERSION);

  /* Initialize timing variable used in rate limit */
  if (maxrate > 0)
    etimeLastPkt = sl_dtime ();

  /* Set SeedLink connection to non-blocking */
  sl_set_blockingmode (slconn, 1);

  /* Loop with the connection manager */
  while ((retval = sl_collect (slconn, &packetinfo,
                               plbuffer, (uint32_t)sizeof (plbuffer))) != SLTERMINATE)
  {
    /* Check if we are being asked to terminate */
    if (!slconn->terminate &&
        (tport_getflag (&regionOut) == TERMINATE ||
         tport_getflag (&regionOut) == myPid))
    {
      logit ("t", "slink2ew terminating on request\n");
      sl_terminate (slconn);
    }

    /* Check if we need to send heartbeat message */
    if (time (&timeNow) - timeLastBeat >= heartbeatInt)
    {
      timeLastBeat = timeNow;
      report_status (&hrtLogo, 0, "");
    }

    /* Check if a packet arrived */
    if (retval == SLPACKET)
    {
      packet_handler (packetinfo, plbuffer, packetinfo->payloadcollected);

      /* Save intermediate state file */
      if (statefile && stateint)
      {
        if (++packetcnt >= stateint)
        {
          sl_savestate (slconn, statefile);
          packetcnt = 0;
        }
      }
      /* Limit processing speed to specified maximum data rate */
      /* warning: rudimentary trapping for overflow errors */
      if (maxrate > 0)
      {
        long usec;

        etimeThisPkt = sl_dtime ();

        /* Minimum microsec for packet size and maxrate */
        usec = (long)(packetinfo->payloadcollected * 1.0e6 / maxrate);

        /* Subtract interval since last packet */
        usec -= (long)((etimeThisPkt - etimeLastPkt) * 1.0e6);

        if (usec > 0 && usec < 1e6)
        { /* Pad interval to limit rate */
          sl_usleep (usec);
        }

        etimeLastPkt = sl_dtime ();
      }

      prethrottle = 0;
    }
    else if (retval == SLNOPACKET) /* If no data, throttle the loop */
    {
      /* Throttling scheme:
       *
       * First let the loop query the non-blocking socket 10 times (prethrottle),
       * this should allow the TCP subsystem time enough when data has reached the
       * host and is in the stack buffers.
       *
       * Next progressively throttle the loop by sleeping for 5ms and increasing
       * the sleep time by 5ms increments until a maximum of 100ms.  Continue to
       * throttle the loop by sleeping for 100ms until data arrives.
       */
      prethrottle++;

      if (prethrottle >= 10)
      {
        if (throttlewait < 100 || throttlecnt == 1)
          throttlewait = 5 * throttlecnt;

        throttlecnt++;

        sleep_ew (throttlewait);
      }
      else
      {
        throttlecnt = 1;
      }
    }
    else /* An unmanageable return value, cannot continue */
    {
      if (retval == SLTOOLARGE)
        logit ("et", "slink2ew: Error - packet too large for receiving buffer\n");
      else
        logit ("et", "slink2ew: Error - unknown return from sl_collect()\n");

      logit ("et", "slink2ew: Fatal error - cannot continue\n");
      break;
    }
  }

  /* Make sure everything is shut down and save the state file */
  sl_disconnect (slconn);

  if (statefile)
    sl_savestate (slconn, statefile);

  tport_detach (&regionOut);
  if (mregionOutP == &mregionOut)
    tport_detach (&mregionOut);

  logit ("t", "%s terminated\n", argv[0]);

  return 0;
} /* End of main() */

// message callback handler
// msg - the message data
// msglength - the message length
// version - the message version (MSG_TBUF, MSG_MSEED2, MSG_MSEED3)
// returns 0 if successful
int
message_handler (char *msg, int msglength, char version)
{
  switch (version)
  {
  case MSG_TBUF:
    waveLogo.type = typeTracebuf2;
    if (tport_putmsg (&regionOut, &waveLogo,
                      msglength, msg) != PUT_OK)
    {
      logit ("et", "slink2ew: Error sending message via transport.\n");
      return 1;
    }
    if (verbose)
      logit ("t", "slink2ew: sent TBUF message.\n");
    break;
  case MSG_MSEED2:
    waveLogo.type = typeMseed;
    if (tport_putmsg (mregionOutP == NULL ? &regionOut : mregionOutP,
                      &waveLogo, msglength, msg) != PUT_OK)
    {
      logit ("et", "slink2ew: Error sending MSEED message via transport.\n");
      return 2;
    }
    if (verbose)
      logit ("t", "slink2ew: sent MSEED message.\n");
    break;
  default:
    logit ("et", "slink2ew: Error - unsupported MSEED format version %d\n",
           version);
    return 3;
  }
  return 0;
}

/***************************************************************************
 * packet_handler():
 * Process a received packet based on packet type.
 ***************************************************************************/
void
packet_handler (const SLpacketinfo *packetinfo,
                char *payload, uint32_t payloadlength)
{
  unsigned long long int seqnum = packetinfo->seqnum;

  if (verbose >= 2)
    logit ("t", "seq %llu, Received %u bytes of payload format %s\n",
           seqnum, payloadlength,
           sl_formatstr (packetinfo->payloadformat, packetinfo->payloadsubformat));

  /* Process packet, but continue regardless of success or error */
  if (packetinfo->payloadformat == SLPAYLOAD_MSEED2 ||
      packetinfo->payloadformat == SLPAYLOAD_MSEED3)
  {
    mseedprocess (payload, payloadlength, MaxLatency,
                  (writemseed || mregionOutP != NULL) ? message_handler : NULL,
                  !writemseed ? message_handler : NULL);
  }
  else
  {
    logit ("et", "slink2ew: Error - unhandled payload format: %s\n",
           sl_formatstr (packetinfo->payloadformat, packetinfo->payloadsubformat));
  }
} /* End of packet_handler() */

/***************************************************************************
 * report_status():
 * Send error and hearbeat messages to transport ring.
 *
 ***************************************************************************/
void
report_status (MSG_LOGO *pLogo, short code, char *message)
{
  char outMsg[100]; /* The outgoing message.        */
  time_t msgTime;   /* Time of the message.         */

  /*  Get the time of the message */
  time (&msgTime);

  /* Build & process the message based on the type */
  if (pLogo->type == typeHeartbeat)
  {
    sprintf (outMsg, "%ld %ld\n", (long)msgTime, (long)myPid);

    /* Write the message to the output region */
    if (tport_putmsg (&regionOut, &hrtLogo, (long)strlen (outMsg),
                      outMsg) != PUT_OK)
    {
      /* Log an error message */
      logit ("et", "slink2ew: Failed to send a heartbeat message (%d).\n",
             code);
    }
  }
  else
  {
    if (message)
    {
      sprintf (outMsg, "%ld %hd %s\n", (long)msgTime, code, message);
      logit ("t", "Error:%d (%s)\n", code, message);
    }
    else
    {
      sprintf (outMsg, "%ld %hd\n", (long)msgTime, code);
      logit ("t", "Error:%d (No description)\n", code);
    }

    /* Write the message to the output region  */
    if (tport_putmsg (&regionOut, &errLogo, (long)strlen (outMsg),
                      outMsg) != PUT_OK)
    {
      /* Log an error message */
      logit ("et", "slink2ew: Failed to send an error message (%d).\n",
             code);
    }
  }
} /* End of report_status() */

/***************************************************************************
 * auth_value:
 *
 * A callback function returning the auth_buffer, which is a string to
 * be sumitted with the SeedLink AUTH command.
 *
 * Returns authorization value string on success, and NULL on failure
 ***************************************************************************/
const char *
auth_value (const char *server, void *data)
{
  (void)server; /* Server name is not used in this case */
  (void)data;   /* User-supplied data is not used in this case */

  return auth_buffer;
}

/***************************************************************************
 * configure():
 * Process configuration parameters.
 *
 ***************************************************************************/
void
configure (char **argvec)
{

  /* Initialize name of log-file & open it */
  logit_init (argvec[1], 0, 512, 1);

  /* Read module config file */
  if (proc_configfile (argvec[1]) == EW_FAILURE)
  {
    fprintf (stderr, "%s: configure() failed \n", argvec[0]);
    exit (EW_FAILURE);
  }

  /* Read node configuration info */
  if (!myInstIdConf)
  {
    if (GetLocalInst (&myInstId) != 0)
    {
      fprintf (stderr, "slink2ew: Error getting myInstId.\n");
      exit (EW_FAILURE);
    }
    myInstIdConf = 1;
  }

  /* Lookup the ring keys */
  if ((ringKey = GetKey (ringName)) == -1)
  {
    fprintf (stderr, "slink2ew:  Invalid ring name <%s>; exiting!\n", ringName);
    exit (EW_FAILURE);
  }
  if (mseedRingName[0] && ((mseedRingKey = GetKey (mseedRingName)) == -1))
  {
    fprintf (stderr, "slink2ew:  Invalid mseed ring name <%s>; exiting!\n", mseedRingName);
    exit (EW_FAILURE);
  }

  /* Look up message types of interest */
  if (GetType ("TYPE_HEARTBEAT", &typeHeartbeat) != 0)
  {
    fprintf (stderr, "slink2ew: Invalid message type <TYPE_HEARTBEAT>; exiting!\n");
    exit (EW_FAILURE);
  }
  if (GetType ("TYPE_ERROR", &typeError) != 0)
  {
    fprintf (stderr, "slink2ew: Invalid message type <TYPE_ERROR>; exiting!\n");
    exit (EW_FAILURE);
  }
  if (GetType ("TYPE_TRACEBUF2", &typeTracebuf2) != 0)
  {
    fprintf (stderr, "slink2ew: Invalid message type <TYPE_TRACEBUF2>; exiting!\n");
    exit (EW_FAILURE);
  }
  if (writemseed || mseedRingName[0])
  {
    if (GetType ("TYPE_MSEED", &typeMseed) != 0)
    {
      fprintf (stderr, "slink2ew: Invalid message type <TYPE_MSEED>; exiting!\n");
      exit (EW_FAILURE);
    }
  }

  /* Set up logos for outgoing messages */
  hrtLogo.instid = myInstId;
  hrtLogo.mod    = myModId;
  hrtLogo.type   = typeHeartbeat;

  errLogo.instid = myInstId;
  errLogo.mod    = myModId;
  errLogo.type   = typeError;

  waveLogo.instid = myInstId;
  waveLogo.mod    = myModId;
  waveLogo.type   = 0; /* This gets set to the appropriate type later */

  /* Get my process ID so I can let statmgr restart me */
  myPid = getpid ();

  logit ("et", "%s(%s): Read command file <%s>\n",
         argvec[0], myModName, argvec[1]);

  /* Initialize libslink logging facility */
  sl_loginit (verbose,
              &logit_msg, "",
              &logit_err, "error: ");

  /* Attempt to recover sequence numbers from state file */
  if (statefile)
  {
    if (sl_recoverstate (slconn, statefile) < 0)
    {
      sl_log (1, 0, "state recovery failed\n");
    }
  }

  /* Reinitialize the logging level */
  logit_init (argvec[1], 0, 512, logSwitch);

  return;
} /* End of configure() */

/***************************************************************************
 * proc_configfile():
 * Process the module configuration parameters.
 *
 ***************************************************************************/
int
proc_configfile (char *configfile)
{
  char *com;
  char *str;
  int nfiles;
  int success;

  int networktimeout = 0;
  int networkdelay   = 0;
  int keepalive      = 0;

  char server_address[200] = {0};

  char slhost[150] = {0};
  char slport[50]  = {0};
  char *selectors  = NULL;
  char *paramdir   = NULL;

  /* Some important initial values or defaults */
  ringName[0]      = '\0';
  myModName[0]     = '\0';
  heartbeatInt     = -1;
  logSwitch        = -1;
  mseedRingName[0] = '\0';

  /* Open the main configuration file */
  nfiles = k_open (configfile);
  if (nfiles == 0)
  {
    fprintf (stderr,
             "slink2ew: Error opening command file <%s>; exiting!\n",
             configfile);
    return EW_FAILURE;
  }

  /* Process all command files */
  while (nfiles > 0) /* While there are command files open */
  {
    while (k_rd ()) /* Read next line from active file  */
    {
      com = k_str (); /* Get the first token from line */

      /* Ignore blank lines & comments */
      if (!com)
        continue;
      if (com[0] == '#')
        continue;

      /* Open a nested configuration file */
      if (com[0] == '@')
      {
        success = nfiles + 1;
        nfiles  = k_open (&com[1]);
        if (nfiles != success)
        {
          fprintf (stderr,
                   "slink2ew: Error opening command file <%s>; exiting!\n",
                   &com[1]);
          return EW_FAILURE;
        }
        continue;
      }

      /* Process anything else as a command */
      if (k_its ("MyModuleId"))
      {
        if ((str = k_str ()))
        {
          if (strlen (str) >= sizeof (myModName))
          {
            fprintf (stderr, "MyModId too long; max is %zu\n", sizeof (myModName) - 1);
            return EW_FAILURE;
          }

          strcpy (myModName, str);

          /* Lookup module ID */
          if (GetModId (myModName, &myModId) != 0)
          {
            fprintf (stderr, "slink2ew: Error getting module ID for %s\n", myModName);
            return EW_FAILURE;
          }
        }
      }

      else if (k_its ("RingName"))
      {
        if ((str = k_str ()))
        {
          if (strlen (str) >= sizeof (ringName))
          {
            fprintf (stderr, "OutRing name too long; max is %zu\n", sizeof (ringName) - 1);
            return EW_FAILURE;
          }

          strcpy (ringName, str);
        }
      }

      else if (k_its ("HeartBeatInterval"))
      {
        heartbeatInt = k_long ();
      }

      else if (k_its ("LogFile"))
      {
        logSwitch = k_int ();
      }

      else if (k_its ("Verbosity"))
      {
        if ((str = k_str ()))
        {
          verbose = atoi (str);
        }
        if (!str)
        {
          fprintf (stderr, "Verbosity is unspecified, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("SLhost"))
      {
        if ((str = k_str ()))
        {
          if (strlen (str) >= sizeof (slhost))
          {
            fprintf (stderr, "SLaddr too long; max is %zu characters\n", sizeof (slhost));
            return EW_FAILURE;
          }
          strncpy (slhost, str, sizeof (slhost) - 1);
        }
      }

      else if (k_its ("SLport"))
      {
        if ((str = k_str ()))

        {
          if (strlen (str) >= sizeof (slport))
          {
            fprintf (stderr, "SLport too long; max is %zu characters\n", sizeof (slport));
            return EW_FAILURE;
          }
          strncpy (slport, str, sizeof (slport) - 1);
        }
      }

      else if (k_its ("StateFile"))
      {
        str = k_str ();
        if (str && str[0] != '#')
        {
          statefile = strdup (str);
        }
        else
        {
          /* Build a state file name of the form 'slink<modid>.state' and
             prepend the parameter directory */
          paramdir  = getenv ("EW_PARAMS");
          statefile = (char *)malloc (strlen (paramdir) + 32);

#ifdef _WINNT
          sprintf (statefile, "%s\\slink%d.state", paramdir, myModId);
#else /* *nix brand */
          sprintf (statefile, "%s/slink%d.state", paramdir, myModId);
#endif
        }
      }

      else if (k_its ("StateFileInt"))
      {
        if ((str = k_str ()))
        {
          stateint = atoi (str);
        }
        if (!str || stateint < 0)
        {
          fprintf (stderr, "StateFileInt is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("NetworkTimeout"))
      {
        if ((str = k_str ()))
        {
          networktimeout = atoi (str);
        }
        if (!str || networktimeout < 0)
        {
          fprintf (stderr, "NetworkTimeout is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }

        sl_set_idletimeout (slconn, networktimeout);
      }

      else if (k_its ("NetworkDelay"))
      {
        if ((str = k_str ()))
        {
          networkdelay = atoi (str);
        }
        if (!str || networkdelay < 0)
        {
          fprintf (stderr, "NetworkDelay is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }

        sl_set_reconnectdelay (slconn, networkdelay);
      }

      else if (k_its ("KeepAlive"))
      {
        if ((str = k_str ()))
        {
          keepalive = atoi (str);
        }
        if (!str || keepalive < 0)
        {
          fprintf (stderr, "KeepAlive is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }

        sl_set_keepalive (slconn, keepalive);
      }

      else if (k_its ("MaxRate"))
      {
        if ((str = k_str ()))
        {
          maxrate = atoi (str);
        }
        if (!str || maxrate < 0)
        {
          fprintf (stderr, "MaxRate is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("SLRecSize"))
      {
        fprintf (stderr, "slink2ew: parameter SLRecSize is not longer needed; ignoring\n");
      }

      else if (k_its ("WriteMSEED"))
      {
        writemseed = k_int ();
      }

      else if (k_its ("MseedRing"))
      {
        if ((str = k_str ()))
        {
          if (strlen (str) >= sizeof (mseedRingName))
          {
            fprintf (stderr, "Mseed ring name too long; max is %zu characters\n", sizeof (mseedRingName) - 1);
            return EW_FAILURE;
          }
          strcpy (mseedRingName, str);
        }
      }

      else if (k_its ("instId"))
      {
        if ((str = k_str ()))
        {
          if (GetInst (str, &myInstId))
          {
            fprintf (stderr, "slink2ew: Unknown installation ID '%s'\n", str);
            return EW_FAILURE;
          }

          myInstIdConf = 1;
        }
      }

      else if (k_its ("Selectors"))
      {
        if ((str = k_str ()))
        {
          if (strlen (str) >= 100)
          {
            fprintf (stderr, "Selectors too long; max is 100 characters\n");
            return EW_FAILURE;
          }
          selectors = strdup (str);
        }
      }

      else if (k_its ("Stream"))
      {
        if ((str = k_str ()))
        {
          char stationid[22]     = {0};
          char streamselect[200] = {0};

          strncpy (stationid, str, sizeof (stationid) - 1);

          /* Build a selector list from an optional 3rd value */
          if ((str = k_str ()))
          {
            strncpy (streamselect, str, sizeof (streamselect) - 1);
          }
          else
          {
            k_err (); /* Clear the error if there was no selectors */
          }

          if (streamselect[0] != '\0')
            sl_add_stream (slconn, stationid, streamselect, SL_UNSETSEQUENCE, NULL);
          else
            sl_add_stream (slconn, stationid, selectors, SL_UNSETSEQUENCE, NULL);
        }
      }
      /* optional setting to reject packets for EW with latency greater than this value */
      else if (k_its ("MaxLatency"))
      {
        if ((str = k_str ()))
        {
          MaxLatency = (double)atoi (str);
        }
        if (!str || MaxLatency < 0)
        {
          fprintf (stderr, "MaxLatency is unspecified or negative, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("TLS"))
      {
        if ((str = k_str ()))
        {
          slconn->tls = atoi (str);
        }
        if (!str)
        {
          fprintf (stderr, "TLS is unspecified, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("TLS_CA_File"))
      {
        int status;

        if ((str = k_str ()))
        {
          status = setenv ("LIBSLINK_CA_CERT_FILE", str, 1);
        }
        if (!str || status)
        {
          fprintf (stderr, "TLS_CA_File is unspecified or error setting, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("TLS_CA_Path"))
      {
        int status;

        if ((str = k_str ()))
        {
          status = setenv ("LIBSLINK_CA_CERT_PATH", str, 1);
        }
        if (!str || status)
        {
          fprintf (stderr, "TLS_CA_Path is unspecified or error setting, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("TLS_Debug"))
      {
        int status;

        if ((str = k_str ()))
        {
          status = setenv ("LIBSLINK_TLS_DEBUG", str, 1);
        }
        if (!str || status)
        {
          fprintf (stderr, "TLS_Debug is unspecified or error setting, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("TLS_UnverifiedCertOK"))
      {
        int status;

        if ((str = k_str ()))
        {
          status = setenv ("LIBSLINK_CERT_UNVERIFIED_OK", str, 1);
        }
        if (!str || status)
        {
          fprintf (stderr, "TLS_UnverifiedCertOK is unspecified or error setting, quiting\n");
          return EW_FAILURE;
        }
      }

      else if (k_its ("UserPass"))
      {
        char *user = NULL;
        char *pass = NULL;

        user = k_str ();
        pass = k_str ();

        if (user && pass)
        {
          snprintf (auth_buffer, sizeof (auth_buffer), "USERPASS %s %s", user, pass);
          sl_set_auth_params (slconn, auth_value, NULL, NULL);
        }
        else
        {
          fprintf (stderr, "UserPass fields are not both specified, quiting\n");
          return EW_FAILURE;
        }
      }

      /* Unknown command */
      else
      {
        fprintf (stderr, "slink2ew: <%s> Unknown command in <%s>.\n",
                 com, configfile);
        continue;
      }

      /* See if there were any errors processing the command */
      if (k_err ())
      {
        fprintf (stderr,
                 "slink2ew: Bad command in <%s>; exiting!\n\t%s\n",
                 configfile, k_com ());
        return EW_FAILURE;
      }

    } /** while k_rd() **/

    nfiles = k_close ();

  } /** while nfiles **/

  /* Configure the SeedLink server address */
  snprintf (server_address, sizeof (server_address), "%s:%s", slhost, slport);
  sl_set_serveraddress (slconn, server_address);

  /* Configure all-station mode if no streams were specified */
  if (slconn->streams == NULL)
    sl_set_allstation_params (slconn, selectors, SL_UNSETSEQUENCE, NULL);

  /* Check for required parameters */
  if (myModName[0] == '\0')
  {
    fprintf (stderr, "slink2ew: No MyModId parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }
  if (ringName[0] == '\0')
  {
    fprintf (stderr, "slink2ew: No OutRing parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }
  if (heartbeatInt == -1)
  {
    fprintf (stderr, "slink2ew: No HeartBeatInterval parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }
  if (logSwitch == -1)
  {
    fprintf (stderr, "slink2ew: No LogFile parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }
  if (slhost[0] == '\0')
  {
    fprintf (stderr, "slink2ew: No SLhost parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }
  if (slport[0] == '\0')
  {
    fprintf (stderr, "slink2ew: No SLport parameter found in %s\n",
             configfile);
    return EW_FAILURE;
  }

  free (selectors);

  return EW_SUCCESS;
} /* End of proc_configfile() */

/***************************************************************************
 * logit_msg() and logit_err():
 *
 * Hooks for Earthworm logging facility.  These are used via function
 * pointers by the SeedLink library.
 ***************************************************************************/
static void
logit_msg (const char *msg)
{
  logit ("t", "%s", msg);
}

static void
logit_err (const char *msg)
{
  logit ("et", "%s", msg);
}

/***************************************************************************
 * usage():
 * Print the usage message and exit.
 ***************************************************************************/
static void
usage (const char *progname)
{
  fprintf (stderr, "slink2ew version %s\n", VERSION);
  fprintf (stderr, "Usage: %s <config file>\n", progname);
} /* End of usage() */
