
       /************************************************************
        *                        port_un.c                         *
        *                                                          *
        *  Get Trimble GPS data through a com port.                *
        *  GPS data is assumed to be in TSIP format.               *
        *  The functions in this file are Unix (Linux) compatible. *
        ************************************************************/

#include <errno.h>
#include <sys/types.h>   // for open
#include <sys/stat.h>    // for open
#include <fcntl.h>       // for open
#include <stdint.h>      // for uint8_t
#include <string.h>      // for strerror
#include <termios.h>     // for tcsetattr, tcflush
#include <unistd.h>      // for tcsetattr
#include <earthworm.h>   // for logit
#include "adsendxs.h"    // for DLE, ETX

static int fd;
static int debug = 1;


/***************************************************
  SetInterfaceAttribs

  Set attributes of the COM port.
  speed = com port speed, as defined in termios.h
  ReadFileTimeout = Time out after this many msec
 ***************************************************/

int SetInterfaceAttribs(int fd, int speed, int ReadFileTimeout)
{
    struct termios tty;

    if (tcgetattr(fd, &tty) < 0)
    {
        logit( "et", "Error from tcgetattr: %s\n", strerror(errno) );
        return -1;
    }

    cfsetospeed(&tty, (speed_t)speed);   // Output speed
    cfsetispeed(&tty, (speed_t)speed);   // Input speed

    tty.c_cflag |= (CLOCAL | CREAD);     // ignore modem controls
    tty.c_cflag &= ~CSIZE;
    tty.c_cflag |= CS8;                  // 8-bit characters
    tty.c_cflag &= ~PARENB;              // no parity bit
    tty.c_cflag &= ~CSTOPB;              // only need 1 stop bit
    tty.c_cflag &= ~CRTSCTS;             // no hardware flow control

    /* Setup for non-canonical mode */
    tty.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
    tty.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
    tty.c_oflag &= ~OPOST;

    /* Fetch bytes as they become available */
    tty.c_cc[VMIN] = 1;                        // The read is satisfied after MIN chars
    tty.c_cc[VTIME] = ReadFileTimeout / 100;   // Time out after 0.1*TIME seconds

    if (tcsetattr(fd, TCSANOW, &tty) != 0)
    {
        logit( "et", "Error from tcsetattr: %s\n", strerror(errno) );
        return -1;
    }
    return 0;
}


/*****************************************************
  OpenPort

  Open the com port and set attributes.
 *****************************************************/

int OpenPort( char PortName[], int BaudRate, int ReadFileTimeout )
{
   int speed;

/* Make sure the baud rate is valid.
   Set speed to a constant defined in termios.h
   ********************************************/
   switch ( BaudRate )
   {
      case 1200:
         speed = B1200;
         break;
      case 2400:
         speed = B2400;
         break;
      case 4800:
         speed = B4800;
         break;
      case 9600:
         speed = B9600;
         break;
      case 19200:
         speed = B19200;
         break;
      case 38400:
         speed = B38400;
         break;
      case 57600:
         speed = B57600;
         break;
      default:
         logit( "et", "Invalid baud rate: %d\n", BaudRate );
         return -1;
   }

/* Open the port in read-write mode
   ********************************/
   fd = open( PortName, O_RDWR | O_NOCTTY | O_SYNC );
   if ( fd < 0 )
   {
      logit( "et", "Error opening %s: %s\n", PortName, strerror(errno) );
      return -1;
   }
   if ( debug ) logit( "et", "Port %s opened successfully.\n", PortName );

/* 8 bits, no parity, 1 stop bit
   *****************************/
   if ( SetInterfaceAttribs(fd, speed, ReadFileTimeout) != 0 )
   {
      logit( "et", "Error setting interface attributes.\n" );
      return -1;
   }
   if ( debug ) logit( "et", "Interface attributes set successfully.\n" );
   return 0;
}


/************************************************
  ClosePort

  Close the com port.
 ************************************************/

void ClosePort( void )
{
   close( fd );
   return;
}


/******************************************************
  PurgeComInput

  Purge the input com port buffer of any leftover
  characters from the last read.
 ******************************************************/

int PurgeComInput( void )
{
   if ( tcflush(fd, TCIFLUSH) == -1 )
   {
      logit( "et", "Error from tcflush: %s\n", strerror(errno) );
      return -1;
   }
   return 0;
}


/********************************************************
  ReadGpsPacket

  Read one packet from the GPS receiver via com port.
  The com port is left open on return.
  Returns number of bytes read (could be 0).
  Returns -1 if an error occurred.
 ********************************************************/

int ReadGpsPacket( uint8_t packet[] )
{
   uint8_t chr;
   int     npacketchar = 0;
   int     nDLE = 0;
   int     first = 1;
   int     rdlen;

/* Read a packet from the GPS receiver, one byte at a time.
   The TSIP packet format is:
   <DLE><id><data string bytes><DLE><ETX>
   *******************************************************/
   while ( 1 )
   {
      uint8_t buf;

      rdlen = read( fd, &buf, 1 );

      if ( rdlen < 0 )         // Error reading the character
      {
         logit( "et", "read error in ReadGpsPacket: %d: %s\n", rdlen, strerror(errno) );
         return -1;
      }

      if ( rdlen == 0 )        // The read timed out
      {
         logit( "et", "Error in ReadGpsPacket.  Timed out reading from GPS receiver.\n" );
         return -1;
      }

/* The first character read must be a DLE
   **************************************/
      chr = buf;
      if ( first && (chr != DLE) )
      {
         logit( "et", "ReadGpsPacket error.  First character in packet is not DLE.\n" );
         return -1;
      }
      else
         first = 0;

/* Count the total DLE's received
   ******************************/
      if ( chr == DLE ) nDLE++;
      if ( (chr == ETX) && (nDLE % 2 == 0) ) break;

/* Check for buffer overflow
   *************************/
      if ( npacketchar + 1 > MAXGPSPACKETSIZE )
      {
         logit( "et", "ReadGpsPacket buffer overflow.\n" );
         return -1;
      }

/* Save the byte
   *************/
      packet[npacketchar++] = chr;
   }
   packet[npacketchar++] = chr;   // Save the final ETX character
   return npacketchar;            // Return number of bytes read
}


/************************************************
  WriteGpsPacket

  Write packet to GPS receiver via com port.
  Returns  0 if all ok,
          -1 if a write error occurred
 ************************************************/

int WriteGpsPacket( uint8_t packet[], int nchar )
{
   int wlen;

   wlen = write( fd, packet, nchar );
   if ( wlen < 0 )      // write returned an error
   {
      logit( "et", "Error writing GPS packet.\n" );
      return -1;
   }

   if ( wlen < nchar )
   {
      logit( "et", "Requested write of %d characters to GPS.  Only %d were written.\n",
              nchar, wlen );
      return -1;
   }
   return 0;
}
