#include "libtypes.h"
#include "libclient.h"
#include "libstrucs.h"
#include "libmsgs.h"
#include "libpoc.h"
#include "libsupport.h"
#include "libpoc.h"
#include "libsample.h"
#include "config.h"
#include "commands.h"
#include "callbacks.h"

#include "logging.h"
#include "options.h"
#include "heart.h"
#include "q6602ew.h"
#include "lib660Interface.h"

#include "earthworm.h"
#include "transport.h"
#include "q660_globals.h"
#include "ew_globals.h"

#define MAX_WAIT_STATE_BEFORE_EXIT 240 /* max seconds to sit in WAIT for reg state */
#define MAXWAITLIBSHUTDOWN 10

static unsigned char InstId = 255;
static pid_t MyPid;
static int _exiting = 0;

void q6602ew_die(char *msg) {
  fprintf(stderr, "%s exiting: %s", Q6602EW_NAME, msg);
  q6602ew_cleanupAndExit();
}

void q6602ew_cleanup() {
  int countdown;
  enum tliberr err;
  enum tlibstate lstate;
  if (_exiting) {
    return;
  }
  _exiting = 1;
  if (_q8Connected) {
    // Stop the thread (S/T)
    lib_change_state(context, LIBSTATE_IDLE, LIBERR_CLOSED);
    cfgphase = CFG_IDLE;        // Wait for the shutdown to occur
    countdown = MAXWAITLIBSHUTDOWN;
    while (context && countdown--) {
      // wait for IDLE. write q660 continuity
      lstate = lib_get_state(context, &err, &opstat);
      if (err == LIBERR_INVCTX) {
        // library already closed somehow
        break;
      }
      if (lstate == LIBSTATE_IDLE) {
        // proceed to TERM
	break;
      }
      sleep_ew(1000);
    }

    lib_change_state(context, LIBSTATE_TERM, LIBERR_CLOSED);
    countdown = MAXWAITLIBSHUTDOWN;

    while (context && countdown--) {
      // wait for TERM. write stn continuity
      lstate = lib_get_state(context, &err, &opstat);
      if (err == LIBERR_INVCTX) {
        // library already closed somehow
	break;
      }
      if (lstate == LIBSTATE_TERM) {
        // proceed to shutdown
	break;
      }
      sleep_ew(1000);
    }

    // Now terminate the connection
    stopConnection();
  }
}

void q6602ew_cleanupAndExit(int i) {
  q6602ew_cleanup();
  exit(1);
}

void setuplogo(MSG_LOGO *logo) {
   /* only get the InstId once */
   if (InstId == 255  && GetLocalInst(&InstId) != 0) {
      fprintf( stderr, "%s: Invalid Installation code; exiting!\n", Q6602EW_NAME);
   }
   logo->mod = QModuleId;
   logo->instid = InstId;
}

void setupSignalHandlers() {
#ifdef X86_WIN32
  WSAStartup (0x101, &wdata);
#else
  signal (SIGPIPE, SIG_IGN) ; /* don't panic if LISS client disconnects */
#endif
#ifndef WIN32
  signal(SIGHUP, q6602ew_cleanupAndExit);
  signal(SIGINT, q6602ew_cleanupAndExit);
  signal(SIGQUIT, q6602ew_cleanupAndExit);
  signal(SIGTERM, q6602ew_cleanupAndExit);
#endif
}

int main (int argc, char **argv)
{
  EW_ShutMeDown = 0;

  if (argc<2) {
    usage();
    exit(1);
  }

  // then read config
  handle_opts(argc, argv);
  logMessage("%s version %s\n", Q6602EW_NAME, Q6602EW_VERSION);
  printConfigStructToLog();
  setupSignalHandlers();

  // logo setups
  if ( GetType( "TYPE_HEARTBEAT", &TypeHB ) != 0 ) {
    fprintf( stderr,
	     "%s: Invalid message type <TYPE_HEARTBEAT>\n", Q6602EW_NAME);
    return( -1 );
  }
  if ( GetType( "TYPE_TRACEBUF2", &TypeTrace ) != 0 ) {
    fprintf( stderr,
	     "%s: Invalid message type <TYPE_TRACEBUF>; exiting!\n", Q6602EW_NAME);
    return(-1);
  }
  if ( GetType( "TYPE_MSEED", &TypeMseed ) != 0 ) {
    fprintf( stderr,
	     "%s: Invalid message type <TYPE_MSEED>; exiting!\n", Q6602EW_NAME);
    return(-1);
  }

  MyPid = getpid();
  if (MyPid == -1)
  {
    fprintf( stderr,
	     "%s: Cannot get PID; exiting!\n", Q6602EW_NAME);
    return(-1);
  }
  setuplogo(&EW_DataLogo);
  setuplogo(&EW_OtherLogo);

#ifndef WIN32
  // attach to the shared memory 
  EW_TraceRegion.mid = -1;
  EW_MseedRegion.mid = -1;
#endif

  if (!(testflag & TEST_STANDALONE)) {
    if (gConfig.MseedRingKey) {
      tport_attach(&EW_MseedRegion, gConfig.MseedRingKey);
      EW_pRegion = &EW_MseedRegion;
    }
    if (gConfig.RingKey) {
      tport_attach(&EW_TraceRegion, gConfig.RingKey);
      EW_pRegion = &EW_TraceRegion;
    }

    start_heartbeat();
  }
  logMessage("Region.mid=%ld, DataLogo.type=%d, DataLogo.mod=%d, DataLogo.instid=%d\n",
    EW_TraceRegion.mid, EW_DataLogo.type, EW_DataLogo.mod, EW_DataLogo.instid);

  createStationThread();
  startConnection(FALSE);

  if (!(testflag & TEST_STANDALONE)) {
    int flag;
    int wait_counter = 0;
    time_t lastStatusUpdate = time(NULL);
    while((flag = tport_getflag( EW_pRegion )) != TERMINATE &&
          flag != MyPid) {
      if( (time(NULL) - lastStatusUpdate) >= gConfig.statusinterval ) {
        lastStatusUpdate = time(NULL);
      }
      sleep_ew(1000);
      if (lib660Interface_waitForState(LIBSTATE_WAIT, 1) == 1) {
         wait_counter++;
      } else {
         wait_counter = 0;
      }
      if (wait_counter >= MAX_WAIT_STATE_BEFORE_EXIT) {
         logit("et", "q6602ew: hung in wait state for more than: %d seconds\n", MAX_WAIT_STATE_BEFORE_EXIT);
         break;
      }
    }
    q6602ew_cleanup();
  } else {
    run_loop();
  }
  EW_ShutMeDown = 1;
  return 0;
}
