/**
    \file       net.h
    \author     <a href="mailto:jcg@ipac.caltech.edu">John Good</a>
 */

/**
    \mainpage   libnet: Client/Server Communications Library

    <hr>
    <center><h2>Client/Server Communications Library</h2></center>
    <p>

    <center><b>Table of Contents</b></a>
    <table border=1 cellpadding=2 width=75%>
        <tr><td>
            <center><table><tr><td>
                <ul>
                <li><a href=#description>   General Description</a></li>
                <li><a href=#control>       Library Routines</a></li>
                </ul>
            </td></tr></table></center>
        </td></tr>
    </table>
    </center>

    <a name=description><h2>General Description</h2></a>

    <p>
    Setting up a command/control communications link between two processes on
    different machines is fairly straightforward if the two can be made to
    follow certain rules.  In particular, if we assume line-at-a-time textual
    communications the problem can be wrapped in a very simple set of routines.

    <p>
    That is what is embodied here.  For examples of both client and server
    programs, see the associated test code.

    <p><hr><p>

    <a name=control><h2>Library Routines</h2></a>
    <ul>
    <li>client_connect(char *, int)</li>
    <li>client_send_cmd(int, char *)</li>
    <li>client_recv_cmd(int, char *)</li>
    <li>server_connect(int)</li>
    <li>server_send_cmd(int, char *)</li>
    <li>server_recv_cmd(int, char *)</li>
    </ul>

    \author     <a href="mailto:jcg@ipac.caltech.edu">John Good</a>
*/

#ifndef ISIS_NET_LIB
#define ISIS_NET_LIB

/* definition of status codes */

#define OK 0
#define ERROR -1
#define DONE -2

/* definition of error codes SM_ */

#define SM_HOST 1     /**< Can't get hostent from the name                */
#define SM_PORT 2     /**< Can't get the port number for the service      */
#define SM_PROTO 3    /**< Can't conver tcp to protocol number            */
#define SM_SOCK 4     /**< Can't get a socket allocated                   */
#define SM_CONNECT 5  /**< Can't connect to server                        */
#define SM_BROKENC 6  /**< Connection is broken                           */
#define SM_BIND 7     /**< Server bind to port failed                     */
#define SM_LISTEN 8   /**< Server listen to port failed                   */
#define SM_ACCEPT 9   /**< Server accept connection failed                */
#define SM_MESG 10    /**< Problem with message format                    */
#define SM_SEND 11    /**< Error sending message                          */
#define SM_UNKNOWN 12 /**< Unknown error from write() to socket           */
#define SM_EAGAIN 13  /**< See write() errors.                              */
#define SM_EWOULDBLOCK                                                         \
  14                /**< See write() errors.                              */
#define SM_EBADF 15 /**< See write() errors.                              */
#define SM_EDESTADDRREQ                                                        \
  16                 /**< See write() errors.                              */
#define SM_EDQUOT 17 /**< See write() errors.                              */
#define SM_EFAULT 18 /**< See write() errors.                              */
#define SM_EFBIG 19  /**< See write() errors.                              */
#define SM_EINTR 20  /**< See write() errors.                              */
#define SM_EINVAL 21 /**< See write() errors.                              */
#define SM_EIO 22    /**< See write() errors.                              */
#define SM_ENOSPC 23 /**< See write() errors.                              */
#define SM_EPERM 24  /**< See write() errors.                              */
#define SM_EPIPE 25  /**< See write() errors.                              */
#define SM_ZERO 26   /**< Socket write() returned a count of zero          */

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

extern int smerr, net_debug;

/**
    Connects to a server program on a remote machine

    \return	an identifier for the socket to be used to
                send/recieve data to/from the server program
 */
int client_connect(char const *machine, int port);

/**
    Sends a command string to a server program that has been connected to
    with client_connect(char *, int).

    \return     OK if the string was sent successfully,
                ERROR otherwise.
 */
int client_send_cmd(int sock, char const *cmdin);

/**
    Recieves a response from a server program that has been connected to
    with client_connect(char *, int).

    \return     OK if the string was recieved successfully,
                ERROR otherwise.
 */
int client_recv_cmd(int sock, char *string);

/**
    Connects a server program to a port on the local machine.

    \return     an identifier for the socket to be used to
                send/recieve data to/from clients.
 */
int server_connect(int port);

/**
    Sends a server response to a client program.

    \return     OK if the string was sent successfully,
                ERROR otherwise.
 */
int server_send_cmd(int sock, char const *cmdin);

/**
    Recieves a command from a client program.

    \return     OK if the string was recieved successfully,
                ERROR otherwise.
 */
int server_recv_cmd(int sock, char *string);

#endif /* ISIS_NET_LIB */
