/*
 * report_gm.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <earthworm.h>
#include <transport.h>
#include "chron3.h"
#include "config.h"
#include "compare.h"
#include "gm_data.h"
#include "report_gm.h"
#include "stalist.h"
#include "status.h"
#include "strong_motion_gm.h"
#include "write_report.h"

#define GM_REPORT_VERSION "1.0 - 2024-05-30"

#define MESSAGE_BUF_SIZE 512
#define MESSAGE_MAX_SIZE (MESSAGE_BUF_SIZE - 1)

#define isTerminate() (tport_getflag(&Ewparm.InRegion) == TERMINATE || \
                       tport_getflag(&Ewparm.InRegion) == Ewparm.MyPid)

static STATION       *StaArray = NULL;
static int           Nsta = 0;

void run_loop()
{
   time_t        timeLastBeat = 0;
   time_t        timeNow;
   MSG_LOGO      logo;                    /* Logo of retrieved msg */
   long          msgLen;                  /* Size of retrieved message */
   unsigned char seq;                     /* msg sequence number from tport_copyfrom() */
   char          msg[MESSAGE_BUF_SIZE];   /* message buffer */
   STRONG_MOTION_GM smg;
   int           rc;                      /* Return code from tport_copyfrom() */
   STATION key, *Sta;

   if (Gparm.Debug)
   {
      printf("report_gm: start run loop\n");
   }

   memset(&key, 0, sizeof(key));
   memset(msg, 0, MESSAGE_BUF_SIZE);
   memset(&smg, 0, sizeof(STRONG_MOTION_GM));
   // attach to shared memory rings
   tport_attach(&Ewparm.InRegion, Gparm.InKey);
   logit("", "report_gm: Attached to public memory region: %ld\n", Gparm.InKey);

   while (!isTerminate())
   {
      time(&timeNow); // get current time
      if (Gparm.HeartbeatInt && timeNow -
          timeLastBeat >= Gparm.HeartbeatInt)
      {
         timeLastBeat = timeNow;
         // send heartbeat
         status(timeNow, Ewparm.TypeHeartBeat, 0, "");
      }

      rc = tport_copyfrom(&Ewparm.InRegion, Gparm.GetLogo, (short)Gparm.nGetLogo,
                          &logo, &msgLen, msg, MESSAGE_MAX_SIZE, &seq);
      switch (rc)
      {
      case GET_NOTRACK:
         logit("et", "report_gm: Tracking error (NTRACK_GET exceeded)\n");
         break;
      case GET_MISS_LAPPED:
         logit("et", "report_gm: Missed msgs (lapped on ring) "
               "before i:%d m:%d t:%d seq:%d\n",
               (int)logo.instid, (int)logo.mod, (int)logo.type, (int)seq);
         break;
      case GET_MISS_SEQGAP:
         logit("et", "report_gm: Gap in sequence# before i:%d m:%d t:%d seq:%d\n",
               (int)logo.instid, (int)logo.mod, (int)logo.type, (int)seq);
         break;
      case GET_TOOBIG:
         logit("et", "report_gm: Retrieved msg is too big: i:%d m:%d t:%d len:%ld\n",
               (int)logo.instid, (int)logo.mod, (int)logo.type, msgLen);
         break;
      case GET_NONE:
         // if message was received and time to report
         if (smg.time && isTimeToReport(timeNow))
         {
            write_report(timeNow, Nsta, StaArray);
            smg.time = 0;
            if (*Gparm.OutputCommand)
            {
               int result;
               if (Gparm.Debug)
               {
                  printf("report_gm: system(%s)\n", Gparm.OutputCommand);
               }
               result = system(Gparm.OutputCommand);
               if (Gparm.Debug)
               {
                  printf("report_gm: system(%s): %d\n", Gparm.OutputCommand, result);
               }
            }
         }
         else
         {
            sleep_ew(100);
         }
         break;
      case GET_OK:
         rc = parse_strong_motion_gm(msg, &smg);
         if (rc)
         {
            logit("et", "report_gm: Retrieved msg could not be parsed: i:%d m:%d t:%d len:%ld\n",
                  (int)logo.instid, (int)logo.mod, (int)logo.type, msgLen);
            logit("et", "report_gm: (%s) %d\n", msg, rc);
            break;
         }
         strcpy(key.sta, smg.sta);
         strcpy(key.net, smg.net);
         strcpy(key.loc, smg.loc);
         Sta = FindSta( StaArray, Nsta, &key );
         if ( !Sta )
         {
            logit("et", "report_gm: could not find station (%s.%s.%s.%s)\n",
                  key.sta, smg.chan, key.net, key.loc);
         }
         else
         {
            save_values(timeNow, Sta, smg.chan, smg.time, smg.pga, smg.pgv);
         }
         break;
      }
   }
   if (Gparm.Debug)
   {
      printf("report_gm: end run loop\n");
   }
}

int main (int argc, char **argv)
{
   char          *configfile;      /* Pointer to name of config file */

   if (argc < 2) {
      fprintf(stderr, "Usage: report_gm configfile\n");
      fprintf(stderr, " Version: %s\n", GM_REPORT_VERSION);
      exit(-1);
   }

   configfile = argv[1];
   /* Initialize name of log-file & open it */
   logit_init (configfile, 0, 256, 1);
   logit("", "report_gm Version: %s\n", GM_REPORT_VERSION);
   config_load(configfile);

   // read the station list and return the number of stations found
   if ( GetStaList( &StaArray, &Nsta, &Gparm ) == -1 ) {
      logit( "e", "report_gm: GetStaList() failed. Exiting.\n" );
      exit(-1);
   }
   // Sort the station list by SCNL.
   // The sorted list is required by bsearch in FindSta.
   qsort(StaArray, Nsta, sizeof(STATION), CompareSNL);
   // log the station list
   LogStaList( StaArray, Nsta );

   if (Gparm.Debug)
   {
      printf("report_gm: version %s\n", GM_REPORT_VERSION);
      printf("report_gm: PID=%d\n", Ewparm.MyPid);
   }

   run_loop();
   if (Gparm.Debug)
   {
      printf("report_gm: finished\n");
   }
   return 0;
}
