/*
 *   THIS FILE IS UNDER RCS - DO NOT MODIFY UNLESS YOU HAVE
 *   CHECKED IT OUT USING THE COMMAND CHECKOUT.
 *
 *    $Id: readsubs.c 7390 2018-05-02 01:55:51Z baker $
 *
 *    Revision history:
 *     $Log$
 *     Revision 1.7  2004/05/11 17:49:07  lombard
 *     Added support for location code, TYPE_CARLSTATRIG_SCNL and TYPE_TRIGLIST_SCNL
 *     messages.
 *     Removed OS2 support
 *
 *     Revision 1.6  2003/11/17 18:52:22  friberg
 *     Updated carlsubtrig for larger subnets (like Caltech) than just
 *     50 stations. I also added a more intelligent error reporting for
 *     the readsubs.c code when this limit might be reached.
 *
 *     Revision 1.5  2002/07/26 16:32:38  dietz
 *     logging change: add * before channels that don't count toward
 *     subnet trigger threshold.
 *
 *     Revision 1.4  2001/03/27 21:16:56  cjbryan
 *     *** empty log message ***
 *
 *     Revision 1.3  2001/03/21 23:10:15  cjbryan
 *     added CVO subnet trigger
 *
 *     Revision 1.2  2000/06/10 04:15:47  lombard
 *     Fixed bug that caused crash on zero-length lines in config files.
 *
 *     Revision 1.1  2000/02/14 16:14:42  lucky
 *     Initial revision
 *
 *
 */

/*
 * readsubs.c: Read the subnet configuration from a file.
 *              1) Create a list of subnets based on the input file.
 */

/*******                                                        *********/
/*      Functions defined in this source file                           */
/*******                                                        *********/

/*      Function: ReadSubnets                                           */
/*                                                                      */
/*      Inputs:         Pointer to a string(input filename)             */
/*                      Pointer to a network definition                 */
/*                                                                      */
/*      Outputs:        Updated network definition(above)               */
/*                                                                      */
/*      Returns:        0 on success                                    */
/*                      Error code as defined in carlsubtrig.h on failure       */

/*******                                                        *********/
/*      System Includes                                                 */
/*******                                                        *********/
#include <stdio.h>

#include <stdlib.h> /* malloc                                       */

#include <string.h> /* strcpy, strlen                               */

/*******                                                        *********/
/*      Earthworm Includes                                              */
/*******                                                        *********/
#include "earthworm.h" /* logit                                        */

#include "k_readcfg.h"
#include "strfuncs.h"

/*******                                                        *********/
/*      CarlSubTrig Includes                                            */
/*******                                                        *********/
#include "carlsubtrig.h"

/*******                                                        *********/
/*      Function definitions                                            */
/*******                                                        *********/

typedef struct
{
  NETWORK *csuNet;
  SUBNET *subnet; /* subnet of interest                   */
} PROCESS_DATA;

#define SPACE_DELIMITERS " "
#define DOT_DELIMITERS "."

/*      Function: processCommand                                            */
static int
processCommand (READCFG_DATA *data)
{
  PROCESS_DATA *p;
  NETWORK *csuNet;
  char staCode[TRACE2_STA_LEN];   /* Station code (name). */
  char compCode[TRACE2_CHAN_LEN]; /* Component code.      */
  char netCode[TRACE2_NET_LEN];   /* Network code.        */
  char locCode[TRACE2_LOC_LEN];   /* Location code        */
  char staList[MAXLINELEN];       /* Stations remaining to be     */
  /*   parsed in stations string.     */
  STATION *station; /* station of interest                  */
  int retVal = 0;   /* Return value for this function.      */
  int j;
  char triggerableFlag; /* CVO flag for subnet triggers        */
  SUBNET *subnet;       /* subnet of interest                   */
  char *line;
  char *str;
  int instation;

  line = k_get ();
  // ignore if not subset
  if (memcmp (line, SUBNET_PARAM, SUBNET_PARAM_SIZE) != 0)
  {
    return 0;
  }
  // increment line past Subset
  line += SUBNET_PARAM_SIZE;
  p      = (PROCESS_DATA *)data->commandData;
  csuNet = p->csuNet;
  subnet = p->subnet++;
  if (csuNet->csuParam.debug >= 3)
  {
    logit ("", "line: %s\n", line);
  }
  // get subnetCode
  str = strtok (line, SPACE_DELIMITERS);
  if (str == NULL)
  {
    logit ("e", "Unable to parse subnet parameters in %s\n", line);
    return ERR_SUB_READ;
  }
  strcpyn (subnet->subnetCode, str, MAX_SUBNET_LEN);
  if (csuNet->csuParam.debug >= 3)
  {
    logit ("", "subnetCode: %s\n", subnet->subnetCode);
  }
  // get minToTrigger
  str = strtok (NULL, SPACE_DELIMITERS);
  if (str == NULL || !*str)
  {
    logit ("e", "Unable to parse subnet parameters in %s\n", line);
    return ERR_SUB_READ;
  }
  subnet->minToTrigger = atoi (str);
  if (csuNet->csuParam.debug >= 3)
  {
    logit ("", "minToTrigger: %d\n", subnet->minToTrigger);
  }
  // get staList
  str = strtok (NULL, "");
  if (str == NULL)
  {
    logit ("e", "Unable to parse subnet parameters in %s\n", line);
    return ERR_SUB_READ;
  }
  // skip over spaces
  while (*str == ' ')
  {
    str++;
  }
  strcpyn (staList, str, MAXLINELEN);
  if (csuNet->csuParam.debug >= 3)
  {
    logit ("", "staList: %s\n", staList);
  }
  instation     = 0;
  subnet->nStas = 0;
  while (*str)
  {
    if (!instation)
    {
      // not in station, increment nStas and set in station
      subnet->nStas++;
      instation = 1;
    }
    else if (*str == ' ')
    {
      // space, not in station
      instation = 0;
    }
    str++;
  }
  subnet->stations    = calloc (subnet->nStas, sizeof (int));
  subnet->triggerable = calloc (subnet->nStas, sizeof (char));

  /* Changed by Eugene Lublinsky, 3/31/Y2K */
  /* added CVO subnet trigger flag */
  triggerableFlag = 1;

  /* Scan the SCN list */
  str = strtok (staList, DOT_DELIMITERS);
  j   = 0;
  while (str)
  {
    if (*str == '|')
    {
      triggerableFlag = 0; /* divider reached */
      if (csuNet->csuParam.debug >= 3)
      {
        logit ("", "triggerableFlag: %d\n", triggerableFlag);
      }
    }
    else
    {
      // save staCode
      strcpyn (staCode, str, TRACE2_STA_LEN);
      if (csuNet->csuParam.debug >= 3)
      {
        logit ("", "staCode: %s\n", staCode);
      }
      // get and save compCode
      str = strtok (NULL, DOT_DELIMITERS);
      if (str == NULL)
      {
        logit ("e", "Unable to parse SCN list in %s\n", staList);
        return ERR_SUB_READ;
      }
      strcpyn (compCode, str, TRACE2_CHAN_LEN);
      if (csuNet->csuParam.debug >= 3)
      {
        logit ("", "compCode: %s\n", compCode);
      }
      // get and save netCode
      str = strtok (NULL, DOT_DELIMITERS);
      if (str == NULL)
      {
        logit ("e", "Unable to parse SCN list in %s\n", staList);
        return ERR_SUB_READ;
      }
      strcpyn (netCode, str, TRACE2_NET_LEN);
      if (csuNet->csuParam.debug >= 3)
      {
        logit ("", "netCode: %s\n", netCode);
      }
      // get and save locCode
      str = strtok (NULL, SPACE_DELIMITERS);
      if (str == NULL)
      {
        logit ("e", "Unable to parse SCN list in %s\n", staList);
        return ERR_SUB_READ;
      }
      strcpyn (locCode, str, TRACE2_LOC_LEN);
      if (csuNet->csuParam.debug >= 3)
      {
        logit ("", "locCode: %s\n", locCode);
      }

      /* Look for staCode in the list                     */
      station = FindStation (staCode, compCode, netCode,
                             locCode, csuNet);
      if (station)
      {
        if (j >= subnet->nStas)
        {
          logit ("e", "Too many stations for subnet %s; max is %d\n",
                 subnet->subnetCode, subnet->nStas);
          return ERR_SUB_READ;
        }
        /* Using pointer arithmetic, add the station index      */
        /* to the subnet's station array                        */
        subnet->stations[j]    = (int)(station - csuNet->stations);
        subnet->triggerable[j] = triggerableFlag;
        j++;
      }
      else
      {
        /*      Station not found, report it            */
        logit ("e", "Unable to find <%s.%s.%s.%s>, subnet %s\n",
               staCode, compCode, netCode, locCode, subnet->subnetCode);
        retVal = ERR_SUB_READ;
      }
      /* Continue only if no error during station list parsing        */
      if (CT_FAILED (retVal))
        break;
    } /* if '|' */
    str = strtok (NULL, DOT_DELIMITERS);
  }
  return retVal;
}

/*      Function: ReadSubnets                                           */
int
ReadSubnets (NETWORK *csuNet)
{
  int result;
  PROCESS_DATA p;
  READCFG_DATA data;

  p.csuNet = csuNet;
  k_initcfg (&data, &p);

  // allocate the subnet array
  csuNet->subnets = (SUBNET *)calloc (csuNet->nSub, sizeof (SUBNET));
  if (csuNet->subnets == NULL)
  {
    logit ("e", "carlsubtrig: Error allocating subnet memory: %d\n",
           csuNet->nSub);
    return ERR_MALLOC;
  }
  p.subnet = csuNet->subnets;
  // properly initialize all the subnet structures
  InitializeSubnet (csuNet->subnets, csuNet->nSub);
  result = k_readcfg (&data, csuNet->csuParam.subFile, &processCommand);
  if (result == 0)
  {
    // adjust numSubAll if it wasn't set by config file
    if (csuNet->numSubAll == 0)
    {
      csuNet->numSubAll = csuNet->nSub;
    }
    if (csuNet->csuParam.debug)
    {
      int i, j;
      logit ("", " Read %d subnets:\n", csuNet->nSub);
      logit ("", " Channels preceded by * do not count toward subnet trigger threshold\n");
      for (i = 0; i < csuNet->nSub; i++)
      {
        logit ("", "%4d   %-3s   %2d   ",
               i, csuNet->subnets[i].subnetCode,
               csuNet->subnets[i].minToTrigger);
        for (j = 0; j < csuNet->subnets[i].nStas; j++)
        {
          if (!csuNet->subnets[i].triggerable[j])
            logit ("", "*");
          logit ("", "%s.%s.%s.%s ",
                 csuNet->stations[csuNet->subnets[i].stations[j]].staCode,
                 csuNet->stations[csuNet->subnets[i].stations[j]].compCode,
                 csuNet->stations[csuNet->subnets[i].stations[j]].netCode,
                 csuNet->stations[csuNet->subnets[i].stations[j]].locCode);
        }
        logit ("", "\n");
      }
    }
  }
  return result;
}
