/*
 * thresh2trig.c:
 * Takes a thresh message as input and produces a triglist_scnl message
 *
 * ewthresh has similar functionality built in, via the ThreshVotes directive.
 * You should probably only use this module if you need the MaxDistance functionality!
 */

/* needed for strptime in time.h */
#define _XOPEN_SOURCE

#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "earthworm.h"
#include "chron3.h"
#include "kom.h"
#include "transport.h"
#include "site.h"
#include "time_ew.h"

#define NCOMMAND                7 // # of required commands you expect to process
#define MAX_STR               255
#define SHORT_STR              64
#define MEDIUM_STR  (SHORT_STR*2)

#define VERSION "0.3 - 2024-12-05"

/* Time constants
*****************/
/* These variables are set here, but can be changed via
   optional configuration file commands */
static int PreTrigger = 15; /* seconds to save before pick time */
static int Duration = 60;

/* Functions in this source file
 *******************************/
void   thresh2trig_config  ( char * );
void   thresh2trig_lookup  ( void );
void   thresh2trig_status  ( unsigned char, short, char * );
int    thresh2trig_process ( char*, char*, MSG_LOGO );
void   make_datestr( const struct tm*, char * );
void   bldtrig_hyp( char *, MSG_LOGO, MSG_LOGO, const struct tm* );
double distance(double, double, double, double, char);
void   bldtrig_site( char *, int, const struct tm* );

static  SHM_INFO  InRegion;      /* shared memory region to use for input  */
static  SHM_INFO  OutRegion;     /* shared memory region to use for output */

#define   MAXLOGO   2
MSG_LOGO  GetLogo[MAXLOGO];    /* array for requesting module,type,instid */
short     nLogo;

static char ArcMsgBuf[MAX_BYTES_PER_EQ]; /* character string to hold event message
                                        MAX_BYTES_PER_EQ is from earthworm.h */

static char TrigMsgBuf[MAX_TRIG_BYTES]; /* to hold the trigger message / file */

/* Things to read or derive from configuration file
 **************************************************/
static char    InRingName[MAX_RING_STR];       /* name of transport ring for input  */
static char    OutRingName[MAX_RING_STR];      /* name of transport ring for output */
static char    MyModName[MAX_MOD_STR];        /* speak as this module name/id      */
static int     LogSwitch;            /* 0 if no logfile should be written */
static long    HeartBeatInterval;    /* seconds between heartbeats        */
static double  MaxDistance = 0;
static char    StationListPath[MAX_STR] = {0};
static int     Debug = 0;
static char    EventIdFile[MAX_DIR_LEN] = "thresh_id.d"; /* file for storing next valid eventid */
static long    NextId = 0;

/* Things to look up in the earthworm.h tables with getutil.c functions
 **********************************************************************/
static long          InRingKey;      /* key of transport ring for input   */
static long          OutRingKey;     /* key of transport ring for output  */
static unsigned char InstId;         /* local installation id             */
static unsigned char MyModId;        /* Module Id for this program        */
static unsigned char TypeHeartBeat;
static unsigned char TypeError;
static unsigned char TypeThreshAlert;
static unsigned char TypeTrigList;

/* Error messages used by thresh2trig
 *********************************/
#define  ERR_MISSMSG       0   /* message missed in transport ring       */
#define  ERR_TOOBIG        1   /* retreived msg too large for buffer     */
#define  ERR_NOTRACK       2   /* msg retreived; tracking limit exceeded */
#define  ERR_FILEIO        3   /* error opening trigger file             */
#define  ERR_DECODETHRESH  4   /* error reading input thresh message    */
static char  Text[150];        /* string for log/error messages          */

pid_t	MyPid; 	/** Out process id is sent with heartbeat **/

/** length of string required by make_datestr  **/
#define	DATESTR_LEN		22	

int main( int argc, char **argv )
{
   time_t      timeNow;          /* current time                  */
   time_t      timeLastBeat;     /* time last heartbeat was sent  */
   long      recsize;          /* size of retrieved message     */
   MSG_LOGO  reclogo;          /* logo of retrieved message     */
   int       res;


/* Check command line arguments
 ******************************/
   if ( argc != 2 )
   {
        fprintf( stderr, "Usage: thresh2trig <configfile>\n" );
        fprintf( stderr, "Version: %s\n", VERSION );
        exit( 0 );
   }

/* Initialize name of log-file & open it
 ***************************************/
   logit_init( argv[1], 0, 256, 1 );

/* Read the configuration file(s)
 ********************************/
   thresh2trig_config( argv[1] );
   logit( "" , "%s: Read command file <%s>\n", argv[0], argv[1] );
   logit( "" , "%s: Version %s\n", argv[0], VERSION);

/* Look up important info from earthworm.h tables
 ************************************************/
   thresh2trig_lookup();

/* Reinitialize logit logging to the desired level
 *************************************************/
   logit_init( argv[1], 0, 256, LogSwitch );
   

/* Get our own process ID for restart purposes
 **********************************************/

   if ((MyPid = getpid ()) == -1)
   {
      logit ("e", "thresh2trig: Call to getpid failed. Exiting.\n");
      exit (-1);
   }

/* Attach to Input shared memory ring
 ************************************/
   tport_attach( &InRegion, InRingKey );
   logit( "", "thresh2trig: Attached to public memory region %s: %ld\n",
          InRingName, InRingKey );

/* Attach to Output shared memory ring
 *************************************/
   tport_attach( &OutRegion, OutRingKey );
   logit( "", "thresh2trig: Attached to public memory region %s: %ld\n",
          OutRingName, OutRingKey );

/* Force a heartbeat to be issued in first pass thru main loop
 *************************************************************/
   timeLastBeat = time(&timeNow) - HeartBeatInterval - 1;

/* Flush the incomming transport ring on startup
 **************************************************/

	while (tport_getmsg (&InRegion, GetLogo, nLogo,  &reclogo,
				&recsize, ArcMsgBuf, sizeof(ArcMsgBuf)- 1) != GET_NONE)
		;

/*----------------------- setup done; start main loop -------------------------*/

   while(1)
   {
     /* send thresh2trig's heartbeat
      ***************************/
        if  ( time(&timeNow) - timeLastBeat  >=  HeartBeatInterval )
        {
            timeLastBeat = timeNow;
            thresh2trig_status( TypeHeartBeat, 0, "" );
        }

     /* Process all new messages
      **************************/
        do
        {
        /* see if a termination has been requested
         *****************************************/
           if ( tport_getflag( &InRegion ) == TERMINATE ||
                tport_getflag( &InRegion ) == MyPid )
           {
           /* detach from shared memory */
                tport_detach( &InRegion );
           /* write a termination msg to log file */
                logit( "t", "thresh2trig: Termination requested; exiting!\n" );
                fflush( stdout );
                exit( 0 );
           }

        /* Get msg & check the return code from transport
         ************************************************/
           res = tport_getmsg( &InRegion, GetLogo, nLogo,
                               &reclogo, &recsize, ArcMsgBuf, sizeof(ArcMsgBuf)-1 );

           if( res == GET_NONE )          /* no more new messages     */
           {
                break;
           }
           else if( res == GET_TOOBIG )   /* next message was too big */
           {                              /* complain and try again   */
                sprintf(Text,
                        "Retrieved msg[%ld] (i%u m%u t%u) too big for ArcMsgBuf[%ld]",
                        recsize, reclogo.instid, reclogo.mod, reclogo.type,
                        (long)sizeof(ArcMsgBuf)-1 );
                thresh2trig_status( TypeError, ERR_TOOBIG, Text );
                continue;
           }
           else if( res == GET_MISS )     /* got a msg, but missed some */
           {
                sprintf( Text,
                        "Missed msg(s)  i%u m%u t%u  %s.",
                         reclogo.instid, reclogo.mod, reclogo.type, InRingName );
                thresh2trig_status( TypeError, ERR_MISSMSG, Text );
           }
           else if( res == GET_NOTRACK ) /* got a msg, but can't tell */
           {                             /* if any were missed        */
                sprintf( Text,
                         "Msg received (i%u m%u t%u); transport.h NTRACK_GET exceeded",
                          reclogo.instid, reclogo.mod, reclogo.type );
                thresh2trig_status( TypeError, ERR_NOTRACK, Text );
           }

        /* Process the message
         *********************/
           ArcMsgBuf[recsize] = '\0';      /*null terminate the message*/

           if( reclogo.type == TypeThreshAlert )
           {
               int ret;
               ret   = thresh2trig_process( ArcMsgBuf, TrigMsgBuf, reclogo );
               if(ret) thresh2trig_status( TypeError, ERR_DECODETHRESH, "" );
           }

        } while( res != GET_NONE );  /*end of message-processing-loop */

        sleep_ew( 1000 );  /* no more messages; wait for new ones to arrive */

   }
/*-----------------------------end of main loop-------------------------------*/

    if ( Site != NULL )
        free( Site );
}

/******************************************************************************
 *  thresh2trig_config() processes command file(s) using kom.c functions;        *
 *                    exits if any errors are encountered.                    *
 ******************************************************************************/
void thresh2trig_config( char *configfile )
{
   char     init[NCOMMAND];       /* init flags, one byte for each required command */
   int      nmiss;                /* number of required commands that were missed   */
   char    *com;
   char    *str;
   int      nfiles;
   int      success;
   int      i;
   
   Site = NULL;     /* will need to be freed if UseSiteList command present */
   
/* Set to zero one init flag for each required command
 *****************************************************/
   for( i=0; i<NCOMMAND; i++ )  init[i] = 0;
   nLogo = 0;

/* Open the main configuration file
 **********************************/
   nfiles = k_open( configfile );
   if ( nfiles == 0 ) {
        logit( "e",
                "thresh2trig: Error opening command file <%s>; exiting!\n",
                 configfile );
        exit( -1 );
   }

/* Process all command files
 ***************************/
   while(nfiles > 0)   /* While there are command files open */
   {
        while(k_rd())        /* Read next line from active file  */
        {
            com = k_str();         /* Get the first token from line */

        /* Ignore blank lines & comments
         *******************************/
            if( !com )           continue;
            if( com[0] == '#' )  continue;

        /* Open a nested configuration file
         **********************************/
            if( com[0] == '@' ) {
               success = nfiles+1;
               nfiles  = k_open(&com[1]);
               if ( nfiles != success ) {
                  logit( "e",
                          "thresh2trig: Error opening command file <%s>; exiting!\n",
                           &com[1] );
                  exit( -1 );
               }
               continue;
            }

        /* Process anything else as a command
         ************************************/
  /*0*/     if( k_its("LogFile") ) {
                LogSwitch = k_int();
                init[0] = 1;
            }
  /*1*/     else if( k_its("MyModuleId") ) {
                str = k_str();
                if(str) strcpy( MyModName, str );
                init[1] = 1;
            }
  /*2*/     else if( k_its("InRingName") ) {
                str = k_str();
                if(str) strcpy( InRingName, str );
                init[2] = 1;
            }
  /*3*/     else if( k_its("OutRingName") ) {
                str = k_str();
                if(str) strcpy( OutRingName, str );
                init[3] = 1;
            }
  /*4*/     else if( k_its("HeartBeatInterval") ) {
                HeartBeatInterval = k_long();
                init[4] = 1;
            }
  /*optional*/
            else if( k_its("EventIdFile") ) {
                str = k_str();
                if (!str || strlen(str) >= MAX_DIR_LEN) {
                   logit("e", "thresh2trig: Invalid length file name in <EventIdFile> command; exiting!\n");
                   exit( -1 );
                }
                strcpy( EventIdFile, str );

                /* Open eventid file for command-processing */
                   success = nfiles+1;
                   nfiles  = k_open(EventIdFile);
                   if (nfiles != success) {
                      logit("e", "thresh2trig: Error opening command file <%s>; exiting!\n", EventIdFile);
                      exit(-1);
                   }
                   logit("", "thresh2trig: opened command file <%s>\n", EventIdFile);
                   continue;
            }
            else if (k_its("next_id")) {
                NextId = k_long() + 1;
                continue;
            }


         /* Enter installation & module to get event messages from
          ********************************************************/
  /*5*/     else if( k_its("GetEventsFrom") ) {
                if ( nLogo+1 >= MAXLOGO ) {
                    logit( "e",
                            "thresh2trig: Too many <GetEventsFrom> commands in <%s>",
                             configfile );
                    logit( "e", "; max=%d; exiting!\n", (int) MAXLOGO/2 );
                    exit( -1 );
                }
                if( ( str=k_str() ) != NULL ) {
                   if( GetInst( str, &GetLogo[nLogo].instid ) != 0 ) {
                       logit( "e",
                               "thresh2trig: Invalid installation name <%s>", str );
                       logit( "e", " in <GetEventsFrom> cmd; exiting!\n" );
                       exit( -1 );
                   }
                   GetLogo[nLogo+1].instid = GetLogo[nLogo].instid;
                }
                if( ( str=k_str() ) != NULL ) {
                   if( GetModId( str, &GetLogo[nLogo].mod ) != 0 ) {
                       logit( "e",
                               "thresh2trig: Invalid module name <%s>", str );
                       logit( "e", " in <GetEventsFrom> cmd; exiting!\n" );
                       exit( -1 );
                   }
                   GetLogo[nLogo+1].mod = GetLogo[nLogo].mod;
                }
                if( GetType( "TYPE_THRESH_ALERT", &GetLogo[nLogo].type ) != 0 ) {
                    logit( "e",
                               "thresh2trig: Invalid message type <TYPE_THRESH_ALERT>" );
                    logit( "e", "; exiting!\n" );
                    exit( -1 );
                }
                ++nLogo;
                init[5] = 1;
                if ( Debug ) {
                    printf("GetLogo[%d] inst:%d module:%d type:%d\n",
                        nLogo, (int) GetLogo[nLogo].instid,
                               (int) GetLogo[nLogo].mod,
                               (int) GetLogo[nLogo].type );
                    printf("GetLogo[%d] inst:%d module:%d type:%d\n",
                        nLogo+1, (int) GetLogo[nLogo+1].instid,
                               (int) GetLogo[nLogo+1].mod,
                               (int) GetLogo[nLogo+1].type );
                }
            }
  /*optional*/
	    else if( k_its("PreTrigger") ) {
                PreTrigger = k_int();
            }
  /*optional*/
	    else if( k_its("Duration") ) {
                Duration = k_int();
            }
  /*optional*/
	    else if( k_its("MaxDistance") ) {
                MaxDistance = k_val();
            }
  /*optional*/
	    else if( k_its("Debug") ) {
                Debug = 1;
            }
  /*6*/
	    else if( k_its("UseStationFile") ) {
                str = k_str();
                if ( !str ) {
                    logit( "e", "thresh2trig: UseStationFile command missing path to station file; exiting!\n" );
                    exit( -1 );
                }
                strcpy( StationListPath, str );
                site_read ( StationListPath );
                init[6] = 1;
                if ( Debug )
                    logit( "", "thresh2trig: Read %d stations from StationFile %s\n", nSite, StationListPath );
            }

         /* Unknown command
          *****************/
            else {
                logit( "e", "thresh2trig: <%s> Unknown command in <%s>.\n",
                         com, configfile );
                continue;
            }

        /* See if there were any errors processing the command
         *****************************************************/
            if( k_err() ) {
               logit( "e",
                       "thresh2trig: Bad <%s> command in <%s>; exiting!\n",
                        com, configfile );
               exit( -1 );
            }
        }
        nfiles = k_close();
   }

/* After all files are closed, check init flags for missed commands
 ******************************************************************/
   nmiss = 0;
   for ( i=0; i<NCOMMAND; i++ )  if( !init[i] ) nmiss++;
   if ( nmiss ) {
       logit( "e", "thresh2trig: ERROR, no " );
       if ( !init[0] )  logit( "e", "<LogFile> "           );
       if ( !init[1] )  logit( "e", "<MyModuleId> "        );
       if ( !init[2] )  logit( "e", "<InRingName> "        );
       if ( !init[3] )  logit( "e", "<OutRingName> "       );
       if ( !init[4] )  logit( "e", "<HeartBeatInterval> " );
       if ( !init[5] )  logit( "e", "<GetEventsFrom> "     );
       if ( !init[6] )  logit( "e", "<UseStationFile> "     );
       logit( "e", "command(s) in <%s>; exiting!\n", configfile );
       exit( -1 );
   }

   return;
}

/******************************************************************************
 *  thresh2trig_lookup( )   Look up important info from earthworm.h tables       *
 ******************************************************************************/
void thresh2trig_lookup( void )
{
/* Look up keys to shared memory regions
   *************************************/
   if( ( InRingKey = GetKey(InRingName) ) == -1 ) {
        fprintf( stderr,
                "thresh2trig:  Invalid ring name <%s>; exiting!\n", InRingName);
        exit( -1 );
   }

/* Look up keys to shared memory regions
   *************************************/
   if( ( OutRingKey = GetKey(OutRingName) ) == -1 ) {
        fprintf( stderr,
                "thresh2trig:  Invalid ring name <%s>; exiting!\n", OutRingName);
        exit( -1 );
   }

/* Look up installations of interest
   *********************************/
   if ( GetLocalInst( &InstId ) != 0 ) {
      fprintf( stderr,
              "thresh2trig: error getting local installation id; exiting!\n" );
      exit( -1 );
   }

/* Look up modules of interest
   ***************************/
   if ( GetModId( MyModName, &MyModId ) != 0 ) {
      fprintf( stderr,
              "thresh2trig: Invalid module name <%s>; exiting!\n", MyModName );
      exit( -1 );
   }

/* Look up message types of interest
   *********************************/
   if ( GetType( "TYPE_HEARTBEAT", &TypeHeartBeat ) != 0 ) {
      fprintf( stderr,
              "thresh2trig: Invalid message type <TYPE_HEARTBEAT>; exiting!\n" );
      exit( -1 );
   }
   if ( GetType( "TYPE_ERROR", &TypeError ) != 0 ) {
      fprintf( stderr,
              "thresh2trig: Invalid message type <TYPE_ERROR>; exiting!\n" );
      exit( -1 );
   }
   if ( GetType( "TYPE_THRESH_ALERT", &TypeThreshAlert ) != 0 ) {
      fprintf( stderr,
              "thresh2trig: Invalid message type <TYPE_THRESH_ALERT>; exiting!\n" );
      exit( -1 );
   }
   if ( GetType( "TYPE_TRIGLIST_SCNL", &TypeTrigList ) != 0 ) {      /* changed for v7.0, D Scott */
      fprintf( stderr,
              "thresh2trig: Invalid message type <TYPE_TRIGLIST_SCNL>; exiting!\n" );
      exit( -1 );
   }

   return;
}

/******************************************************************************
 * thresh2trig_status() builds a heartbeat or error message & puts it into       *
 *                   shared memory.  Writes errors to log file & screen.      *
 ******************************************************************************/
void thresh2trig_status( unsigned char type, short ierr, char *note )
{
   MSG_LOGO    logo;
   char        msg[256];
   long        size;
   time_t        t;

/* Build the message
 *******************/
   logo.instid = InstId;
   logo.mod    = MyModId;
   logo.type   = type;

   time( &t );

   if( type == TypeHeartBeat )
   {
        sprintf( msg, "%ld %d\n", (long) t, MyPid );
   }
   else if( type == TypeError )
   {
        sprintf( msg, "%ld %hd %s\n", (long) t, ierr, note);
        logit( "et", "thresh2trig: %s\n", note );
   }

   size = (long)strlen( msg );   /* don't include the null byte in the message */

/* Write the message to shared memory
 ************************************/
   if( tport_putmsg( &OutRegion, &logo, size, msg ) != PUT_OK )
   {
        if( type == TypeHeartBeat ) {
           logit("et","thresh2trig:  Error sending heartbeat.\n" );
        }
        else if( type == TypeError ) {
           logit("et","thresh2trig:  Error sending error:%d.\n", ierr );
        }
   }

   return;
}

/*******************************************************************
 *  thresh2trig_process( )  read a thresh alert message        *
 *                       and build a trigger message               *
 *******************************************************************/
int thresh2trig_process( char* threshMsg, char* trigMsg, MSG_LOGO incoming_logo )
{
   MSG_LOGO  outgoing_logo;  /* outgoing logo                         */
   char sta[TRACE2_STA_LEN], net[TRACE2_NET_LEN], chan[TRACE2_CHAN_LEN], loc[TRACE2_LOC_LEN], time[27];
   double thresh, value;
   struct tm timeStruct;
   const SITE* mySite;
   int i;

   /* load outgoing logo
   *********************/
   outgoing_logo.instid = InstId;
   outgoing_logo.mod    = MyModId;
   outgoing_logo.type   = TypeTrigList;

   /* parse the threshold message, including misspelling and an inconvenient date format */
   if (sscanf(threshMsg, "SNCL=%6[^.].%8[^.].%3[^.].%2[^.] Thresh=%lf Value=%lf Time=%26[A-Za-z0-9 :]", sta, net, chan, loc, &thresh, &value, time) != 7) {
     if (Debug)
       logit( "e", "Unparseable THRESH_ALERT message: %s\n", threshMsg);
     return -1;
   }
   {
     time_t crossing;
     if (!strptime(time, "%a %b %d %H:%M:%S %Y", &timeStruct)) {
       if (Debug)
         logit( "e", "Unparseable time string: %s\n", time);
       return -1;
     }
     crossing = timegm(&timeStruct) - PreTrigger;
     gmtime_r(&crossing, &timeStruct);
   }

   if (Debug) logit("et", "Processed threshMsg: %s", threshMsg);
   
   /* Find this station */
   for (i = 0; i < nSite; ++i) {
     if (!strcmp(Site[i].name, sta) &&
         !strcmp(Site[i].comp, chan) &&
         !strcmp(Site[i].net, net) &&
         !strcmp(Site[i].loc, loc))
       break;
   }
   if (i == nSite) {
     if (Debug)
       logit( "e", "SCNL not found: %s.%s.%s.%s\n", sta, chan, net, loc);
     return -1;
   }
   mySite = &Site[i];

   if (Debug) logit("et", "Trigger message created:\n");
   /* write 1st part of trigger msg  */
   bldtrig_hyp( trigMsg, incoming_logo, outgoing_logo, &timeStruct);

   /* Write a phase line for each site within StationListDistance */
   for (i = 0; i < nSite; ++i) {
     if (MaxDistance) {
       const double dist = distance(Site[i].lat, Site[i].lon, mySite->lat, mySite->lon, 'K');
       if (dist > MaxDistance) {
         if (Debug)
           logit( "", "Rejected %s.%s.%s.%s: too far (%lf)\n", Site[i].name, Site[i].comp, Site[i].net, Site[i].loc, dist);
         continue;
       }
     }
     bldtrig_site(trigMsg, i, &timeStruct);
   }
   if (Debug) logit("o", "\n\n");

  /* Write trigger message to output ring */
   if( tport_putmsg( &OutRegion, &outgoing_logo, (long)strlen(trigMsg), trigMsg ) != PUT_OK )
   {
      logit("et","thresh2trig: Error writing trigger message to ring.\n" );
   }

   return 0;
}

/**************************************************************
 * bldtrig_hyp() builds the EVENT line of a trigger message   *
 * Modified for author id by alex 7/10/98                     *
 **************************************************************/
void bldtrig_hyp( char *trigmsg, MSG_LOGO incoming, MSG_LOGO outgoing, const struct tm* time )
{
   char datestr[DATESTR_LEN];
   char msgstr[MAX_STR];
   FILE* idfp;

/* Sample EVENT line for trigger message:
EVENT DETECTED     970729 03:01:13.22 UTC EVENT ID:123456 AUTHOR: asdf:asdf\n
0123456789 123456789 123456789 123456789 123456789 123456789
************************************************************/
   make_datestr( time, datestr );
   /* added v2.0 to trigmsg for EW v7.0, D Scott */
   sprintf( trigmsg, "v2.0 EVENT DETECTED     %s UTC EVENT ID: %ld AUTHOR: %03d%03d%03d"
                        " SUBNET: THRESHOLD\n\n",
			datestr, NextId,
			outgoing.type, outgoing.mod, outgoing.instid );
   logit("o", "%s", trigmsg);
   snprintf( msgstr, MAX_STR,  "Sta/Cmp/Net/Loc   Date   Time                       start save       duration in sec.\n");
   //strcat ( trigmsg, "Sta/Cmp/Net/Loc   Date   Time                       start save       duration in sec.\n" );
   strcat ( trigmsg, msgstr );
   if (Debug) logit("o", "%s", msgstr);
   
   snprintf(msgstr, MAX_STR,  "---------------   ------ ---------------    ------------------------------------------\n");
   strcat ( trigmsg, msgstr);
   if (Debug) logit("o", "%s", msgstr);

   ++NextId;

   /* A new event was created; update event_id file...   950413:ldd. */
   idfp = fopen(EventIdFile, "w");
   if (idfp) {
      fprintf(idfp, "# Next available threshold sequence number:\n");
      fprintf(idfp, "next_id %ld\n", NextId);
      fclose(idfp);
   }
}

/****************************************************************
 * bldtrig_site() builds a site line of a trigger message       *
 ****************************************************************/
void bldtrig_site( char *trigmsg, int i, const struct tm* time )
{
   char str[MEDIUM_STR];
   char timestr[DATESTR_LEN];
   
   make_datestr( time, timestr );
   
   snprintf( str, MEDIUM_STR, " %4s %3s %2s %2s %c %s UTC    save: %s %8d\n",
            Site[i].name, Site[i].comp, Site[i].net, Site[i].loc, 'T',
            timestr, timestr, Duration );

   if ( Debug )
     logit("o", "%s", str);

   strcat( trigmsg, str );
}

/*********************************************************************
 * make_datestr()  takes a time as a struct tm and converts          *
 *                 it into a character string in the form of:        *
 *                   "19880123 12:34:12.00"                          *
 *                 It returns a pointer to the new character string  *
 *                                                                   *
 *    NOTE: this requires an output buffer >=21 characters long      *
 *                                                                   *
 *********************************************************************/

void make_datestr( const struct tm* time, char *datestr )
{
    strftime(datestr, DATESTR_LEN, "%Y%m%d %H:%M:%S.00", time);
}

/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*::  This function converts decimal degrees to radians             :*/
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
double deg2rad(double deg) {
  return (deg * M_PI / 180);
}

/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
/*::  This function converts radians to decimal degrees             :*/
/*:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/
double rad2deg(double rad) {
  return (rad * 180 / M_PI);
}

/******************************************************
 * distance()  Calculate distance between 2 lat/lons, *
 *              in (M)iles, (K)ilometers, or (N)?     *
 ******************************************************/
double distance(double lat1, double lon1, double lat2, double lon2, char unit) {
  double dist;
  const double theta = lon1 - lon2;
  dist = sin(deg2rad(lat1)) * sin(deg2rad(lat2)) + cos(deg2rad(lat1)) * cos(deg2rad(lat2)) * cos(deg2rad(theta));
  dist = acos(dist);
  dist = rad2deg(dist);
  dist = dist * 60 * 1.1515;
  switch(unit) {
    case 'M':
      break;
    case 'K':
      dist = dist * 1.609344;
      break;
    case 'N':
      dist = dist * 0.8684;
      break;
  }
  return (dist);
}
