
'''
cosmos0index -- Add on script for COSMOS0 output

cosmos0index  can be called from any module providing COSMOS0 output.
    In the module configuration "ExtCommandName" variable should be set like this:
    "python cosmos0index.py pdl_evid.db cosmosindex.html"
    where the first parameter is PDL database filename and the second one is output HTML file.
    Script will be called after each event processing, with adding parameters for current QID and generated filenames.
    The result is the HTML file with list of generated COSMOS0 files for each event with link to USGS site for this event.


@author:     Max Kokoulin

'''

import sys
import os
import sqlite3
import logging

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
#from _ast import Try

__all__ = []
__version__ = 0.1
__date__ = '2021-09-13'
__updated__ = '2021-09-13'

DEBUG = 0
TESTRUN = 0
PROFILE = 0
FORMAT = '%(asctime)-15s %(clientip)s %(user)-8s %(message)s'
footer = '</TABLE>\n</BODY>\n</HTML>\n'

class CLIError(Exception):
    '''Generic exception to raise and log different fatal errors.'''
    def __init__(self, msg):
        super(CLIError).__init__(type(self))
        self.msg = "E: %s" % msg
    def __str__(self):
        return self.msg
    def __unicode__(self):
        return self.msg

def main(argv=None): # IGNORE:C0111
    logging.basicConfig(format=FORMAT)
    logger = logging.getLogger('cosmos0index')
    
    '''Command line options.'''

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    program_license = '''%s

  Created by Max Kokoulin on %s.
  Copyright 2021 Instrumental Software Technologies.

  Licensed under the Apache License 2.0
  http://www.apache.org/licenses/LICENSE-2.0

  Distributed on an "AS IS" basis without warranties
  or conditions of any kind, either express or implied.

USAGE
''' % (program_shortdesc, str(__date__))

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-v", "--verbose", action="store_true", help="set verbosity level [default: %(default)s]")
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument("database", help="SQLite PDL database file")
        parser.add_argument("indexfile", help="HTML index file to store COSMOS0 output [default: %(default)s]")
        parser.add_argument("eventID", type=int, help="Earthworm EventID")
        parser.add_argument("cosmos0files", help="Semicolon-separated list of full paths to generated COSMOS0 data files")
    
        # Process arguments
        args = parser.parse_args()
    
        if args.verbose:
            print("Verbose mode on")
        
        if not os.path.isfile(args.indexfile):
            f = open(args.indexfile, "wt")
            f.write('<!DOCTYPE HTML>\n<HTML lang=en>\n<HEAD>\n<TITLE></TITLE>\n<meta charset="windows-1252">\n</HEAD>\n<BODY>\n'+
                    '<TABLE id="EventListTable" cellspacing="1" cellpadding="3">\n'+
                    '<TR bgcolor="000060">\n'+
                    '<TH><font size="1" face="Sans-serif" color="FFFFFF">Earthworm QID</font></TH>\n'+
                    '<TH><font size="1" face="Sans-serif" color="FFFFFF">USGS Event ID</font></TH>\n'+
                    '<TH><font size="1" face="Sans-serif" color="FFFFFF">Timestamp</font></TH>\n'+
                    '<TH><font size="1" face="Sans-serif" color="FFFFFF">Magnitude</font></TH>\n'+
                    '<TH><font size="1" face="Sans-serif" color="FFFFFF">COSMOS0 files</font></TH>\n</TR>\n')
            f.write(footer)
            f.close()
        
        try:
            f = open(args.indexfile, "r+b")
            
            conn = sqlite3.connect(args.database)
            cursor = conn.cursor()
            cursor.execute("SELECT * from Events where rowid=(?)", (args.eventID,))
            result = cursor.fetchall()
            conn.close()
            
            f.seek(-len(footer), os.SEEK_END)           
            usgsRef = "Not found"
            if len(result)==1:
                usgsRef='<A href="https://earthquake.usgs.gov/earthquakes/eventpage/' + result[0][0].strip() + '/executive">'+ result[0][0].strip() + "</A>"
            
            cosmos0files = args.cosmos0files.split(";")
            f.write(('<TR bgcolor="#DDDDFF">\n<TD><font size="1" face="Sans-serif">' + str(args.eventID) + "</font></TD>\n"+
            '<TD><font size="1" face="Sans-serif">' + usgsRef + "</font></TD>\n"+
            '<TD><font size="1" face="Sans-serif">' + result[0][1].strip() + "</font></TD>\n"+
            '<TD align="center"><font size="1" face="Sans-serif">' + str(result[0][2]) + "</font></TD>\n"+
            '<TD><font size="1" face="Sans-serif"><TABLE>\n').encode('ascii'))
            for cosmos0file in cosmos0files:
                f.write(('<TR><TD><A href="file://' + cosmos0file + '">' + cosmos0file[cosmos0file.rfind(os.path.sep)+1:] + "</A></TD></TR>\n").encode('ascii'))
            f.write(("</TABLE></font></TD>\n</TR>\n").encode('ascii'))
            f.write(footer.encode('ascii'))
            f.close()
        except Exception as e:
            sys.stderr.write(program_name + ": " + repr(e) + "\n")
            return 1
        
        return 0
        
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 0
    except Exception as e:
        if DEBUG or TESTRUN:
            raise(e)
        indent = len(program_name) * " "
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        return 2

if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-h")
        sys.argv.append("-v")
    if TESTRUN:
        import doctest
        doctest.testmod()
    if PROFILE:
        import cProfile
        import pstats
        profile_filename = 'cosmos0index_profile.txt'
        cProfile.run('main()', profile_filename)
        statsfile = open("profile_stats.txt", "wb")
        p = pstats.Stats(profile_filename, stream=statsfile)
        stats = p.strip_dirs().sort_stats('cumulative')
        stats.print_stats()
        statsfile.close()
        sys.exit(0)
    sys.exit(main())

