"""
ewave-plot

app to return waveform windows as either png or mseed
from configured waveserver_V instances

:copyright:
    Mike Hagerty (m.hagerty@isti.com), 2022
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""

import json
import logging
import os
import shutil
import sys

from ewave_plot.libs.cmd_line import parse_cmd_line, required_arg, optional_arg

from . import installation_dir

fname = 'ewave_plot'

defaults = {
            "waveserverHost": ["localhost"],
            "waveserverPort": [16042, 16022],
            "ewaveHost": "127.0.0.1",
            "ewavePort": 5000,
            "plotColor": "black",
            "width": 800,
            "height": 125,
            "dpi": 100,
            "static_folder": "static",
            "logdir": "./log",
            "loglevel": "INFO",
            "logfile": "%s.log" % fname
        }


def main():

    """
    >>>ewave-plot -c config.json --loglevel INFO --logdir log 
    """

    o = []
    #o.append(optional_arg(arg=['--host'], type=str, description='ipaddr',
                        #help='host ipaddr'))
    #o.append(optional_arg(arg=['-p', '--port'], type=int, description='port',
                        #help='host port'))
    # We will automatically get --logfile, --logdir, --loglevel, --configfile
    # Add configfile explicitly so we can also use -c option
    r = []
    r.append(required_arg(arg=['-c', '--configfile'], type=str, description='path/to/json/config'))
    args, cmdparser = parse_cmd_line(required_args=r, optional_args=o)

    # Read settings from passed --configfile:
    with open(args.configfile, 'r') as file:
        settings_json = json.load(file)
    if settings_json['loglevel']:
        settings_json['loglevel'] = settings_json['loglevel'].upper()

    settings = defaults.copy()
    # Override settings with highest --> lowest precedence:
    # 1. cmdline args
    # 2. settings.json
    # 3. defaults
    for key in defaults:
        if getattr(args, key, None):
            settings[key] = getattr(args, key, None)
        elif key in settings_json:
            settings[key] = settings_json[key]

    message_buf = []

    # MTH: static_folder here means the (config'able) local cache
    #      for mseed/png files,
    #      as opposed to the app/static/styles folder which
    #         installs in conda env and user doesn't see

    # Set static_folder from settings.static_folder:
    static_folder = settings.get('static_folder')
    static_folder = os.path.abspath(static_folder)
    if os.path.exists(static_folder):
        message_buf.append("Static_folder:%s exists" % static_folder)
    else:
        message_buf.append("Static_folder:%s DOES NOT exist --> create it" % static_folder)
        os.makedirs(static_folder, exist_ok=False)

    # Copy the package app/static/styles folder to the local
    #   static_folder so templates find styles:

    styles_folder = os.path.join(static_folder, 'styles')
    if not os.path.exists(styles_folder):
        #print("installation_dir:%s" % installation_dir())
        src = os.path.join(installation_dir(), 'app/static/styles')
        dst = os.path.join(static_folder, 'styles')
        #print("copy from src:%s to dst:%s" % (src, dst))
        #shutil.copytree(src, dst, symlinks=False, ignore=None, dirs_exist_ok=False)
        shutil.copytree(src, dst)

    # Use updated settings dict to set Flask app config

    waveserver_host = settings.get('waveserverHost')
    waveserver_port = settings.get('waveserverPort')

    if not isinstance(waveserver_host, list):
        waveserver_host = [waveserver_host]
    if not isinstance(waveserver_port, list):
        waveserver_port = [waveserver_port]

    if len(waveserver_host) == len(waveserver_port):
        WSV_HOSTS = list(zip(waveserver_host, waveserver_port))
    else:
        WSV_HOSTS = [(host, port) for host in waveserver_host for port in waveserver_port]

    FLASK_PORT = settings.get('ewavePort')
    FLASK_HOST = settings.get('ewaveHost')

    # flask app.config only saves CAPITAL keys

    app_config = {}
    app_config['LOG_DIR'] = settings['logdir']
    app_config['APP_NAME'] = fname
    app_config['APP_LOG_NAME'] = settings['logfile']
    app_config['LOG_LEVEL'] = settings['loglevel']
    app_config['WSV_HOSTS'] = WSV_HOSTS
    app_config['STATIC_FOLDER'] = static_folder
    app_config['FLASK_HOST'] = FLASK_HOST
    app_config['FLASK_PORT'] = FLASK_PORT
    app_config['LOG_TYPE'] = "rotated"

    # MTH: config.py can't be in app/ because if it were,
    #      when we import from it, python reads app/__init__.py
    #      and tries to access Config.STATIC_FOLDER before it is set!
    from ewave_plot.libs.config import Config
    #Config.set_args(**settings_key)
    Config.set_args(**app_config)

    from ewave_plot.app import app

    logger = app.logger
    for msg in message_buf:
        logger.info(msg)

    """
    logger.info("My first log msg from main")
    print("app.static_folder:%s" % app.static_folder)
    print("app.template_folder:%s" % app.template_folder)

    @app.route('/ewave/help')
    def return_usage():
        print("Enter return_usage()")
        return render_template('help.html')
    """

    app.run(debug=False, host=FLASK_HOST, port=FLASK_PORT)

    return

if __name__ == '__main__':
    main()
