"""
   Common functions for logging

   *** Do NOT edit this file directly
       (it will be overwritten by next sync)
   *** Edit master file in:
       .../cmd-log/libs_log.py

   This file gets synced to:
       o aqms-pdl/aqms_pdl/libs/libs_log.py
"""

import os
import sys

import logging

from logging.handlers import TimedRotatingFileHandler

import time

class UtcIsoFormatter(logging.Formatter):

  """Custom logging.Formatter class for logging UTC time in ISO8601 format
  """
  # Use gmtime or universal coordinated time or utc
  #converter = time.gmtime # must not include '()'

  # Format the time for this log event
  def formatTime(self, record, datefmt=None):
    ct = self.converter(record.created)
    if datefmt:
      s = time.strftime(datefmt, ct)
    else:
      # Default UTC ISO8601 format of "YYYY-mm-ddTHH:MM:SS.fffZ"
      t = time.strftime("%Y-%m-%dT%H:%M:%S", ct)
      s = f"{t}.{record.msecs:03.0f}Z"
    return s


def configure_logger(log_funcname=False, **kwargs):
    '''
        levelString = 'INFO' or 'WARN' or 'WARNING', etc
    '''

    logfile = None
    logdir = None
    logutc = False

    # DEFAULTS
    DEFAULT_LEVELSTRING = 'INFO'

    levelString = DEFAULT_LEVELSTRING
    for k,v in kwargs.items():
        if k.lower() in ['log_level', 'loglevel']:
            levelString = v.upper() if v else DEFAULT_LEVELSTRING
        if k.lower() in ['logfile']:
            logfile = v
        if k.lower() in ['log_dir', 'logdir']:
            logdir = v
        if k.lower() in ['log_utc', 'logutc']:
            logutc = True

    fmt = '%(asctime)s [%(levelname)7s] %(message)s'
    if log_funcname:
        fmt = '%(asctime)s [%(levelname)7s] %(funcName)s:%(message)s'
    if logutc:
        fmt = '%(asctime)s UTC [%(levelname)7s] %(message)s'
        if log_funcname:
            fmt = '%(asctime)s UTC [%(levelname)7s] %(funcName)s:%(message)s'

    #formatter = logging.Formatter('%(asctime)s [%(levelname)5s] %(message)s')
    formatter = UtcIsoFormatter(fmt=fmt)

    logger = logging.getLogger()

    logger.level = string_to_logLevel(levelString)

    if logutc:
        formatter.converter = time.gmtime

    if logdir and not logfile:
        #logfile = os.path.basename(sys.argv[0])
        logfile = "%s.log" % os.path.basename(sys.argv[0])

    if logfile is not None:
        if not logdir:
            logdir = os.path.dirname(logfile)

        logfile = os.path.join(logdir, os.path.basename(logfile))

        #print("libs_log logdir=[%s] logfile=[%s]" % (logdir, logfile))

        # if logfile=name.log and logdir was not passed in, then logdir will be empty string
        if logdir and not os.path.exists(logdir):
            try:
                os.makedirs(logdir)
            except:
                raise
        # create file handler 
        fh = TimedRotatingFileHandler(logfile, when="midnight")
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    # MTH: Default logconsole=True if no logfile/logdir set:
    if ('logconsole' in kwargs and kwargs['logconsole']) \
       or logfile is None:
        # create console handler 
        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        logger.addHandler(ch)

    return logger


def string_to_logLevel(levelString):
    '''
       levelString = 'INFO' or 'WARN' or 'WARNING', etc
    '''
    loglevels = {"ERROR": logging.ERROR, "WARN": logging.WARNING,
                 "WARNING": logging.WARNING, "INFO": logging.INFO,
                 "DEBUG": logging.DEBUG, "CRITICAL": logging.CRITICAL,
                }
    if levelString in loglevels:
        return loglevels[levelString]
    else:
        msg = " ".join(list(loglevels.keys()))
        print("ERROR: Unknown --loglevel=%s (not in:{%s})" % (levelString, msg))
        exit(2)


def cleaned_field(field, config_field):

    if '-' in field or '_' in field:
        cleaned = config_field.lower()
    else:
        cleaned = config_field.lower().replace("-","").replace("_","")

    return cleaned


def update_args_with_config(args, config, fname=None):

    if config is None:
        return

    # 1. If args.somefield is None and config['SOMEFIELD'] or config['SOME-FIELD'] or config['SOME_FIELD'],
    #    then update args.somefield from config
    for field, val in vars(args).items():
        #print("args.%s = %s" % (field, val))
        if val is None:
            for c_field in config:
                if cleaned_field(field, c_field) == field:
                    #print("** set args.%s = %s" % (field, config[c_field]))
                    setattr(args, field, config[c_field])
                    break

    if args.logfile is None:
        if fname:
            args.logfile = "%s.log" % fname
        else:
            args.logconsole = True

    if args.logfile and args.logdir:
        args.logfile = os.path.join(args.logdir, args.logfile)

    # 2. If there are other config fields that aren't in cmd line args,
    #    Add them in

    db_fields = ['DB_HOST', 'DB_NAME', 'DB_PORT', 'DB_USER', 'DB_PASSWORD',
                 'conn_string', 'sqlalchemy.url']

    # Any other params in config that *aren't* contained in cmd line args,
    #     add to args *except* for DB connect params that stay with config
    for c_field, c_val in config.items():
        if c_field not in db_fields:
            found = False
            for field, _ in vars(args).items():
                if cleaned_field(field, c_field) == field:
                    found = True
                    break
            if found:
                #print("Already processed config[%s]" % c_field)
                pass
            else:
                new_field = c_field.lower().replace('-','_')
                #print("Set new field: args.%s = config[%s]" % (new_field, c_field))
                setattr(args, new_field, c_val)

    return


# ===========================  read config functions below  =========================
import yaml

from .. import installation_dir

def read_config(requireConfigFile=True, configfile=None, debug=False):

    fname = 'read_config'
    config = None

    #log_msgs = []
    log_msgs = {}
    log_msgs['info'] = []
    log_msgs['warn'] = []
    log_msgs['error'] = []

    # From highest to lowest priority location of config.yml:
    # 1. passed on cmd line --configFile=...
    # 2. os.getcwd() = Location script is called from
    # 3. from env AQMS_DB_CONFIG
    # 4. INSTALL_DIR = Location of yasmine_cli directory
    # 5. None


    if configfile:
        log_msgs['info'].append("%s: configfile passed in:%s" % (fname, configfile))

    if configfile is None:
    # See if it was passed on cmd line:
        if debug:
            print("aqms-pdl read_config() check if configfile was passed on cmd line")
        configfile = pick_off_configFile(sys.argv[1:], debug)
        #if debug and configfile:
        if configfile:
            log_msgs['info'].append("%s: Read from --configfile=[%s]" % (fname, configfile))

    if configfile is None:
        if debug:
            print("aqms-pdl read_config() Look in CWD")
        # Look for config.yml in dir where pdl-to-aqms is called from (changes)
        test_path = os.path.join(os.getcwd(), 'config.yml')
        if os.path.exists(test_path):
            configfile = test_path
        if configfile:
            log_msgs['info'].append("%s: Read from CWD ./config.yml" % (fname))

    if configfile is None:
        if debug:
            print("aqms-pdl read_config() check ENV AQMS_DB_CONFIG")
        configfile = os.getenv('AQMS_DB_CONFIG')
        if configfile:
            log_msgs['info'].append("%s: Read from env var AQMS_DB_CONFIG-->configfile:%s" % (fname, configfile))

    if configfile is None:
        if debug:
            print("aqms-pdl read_config() Look in INSTALL_DIR")
        # Look for config.yml in dir where pdl_to_aqms.py is located (it never changes)
        #SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
        #test_path = os.path.join(SCRIPT_DIR, 'config.yml')
        INSTALL_DIR = installation_dir()
        test_path = os.path.join(INSTALL_DIR, 'config.yml')
        if os.path.exists(test_path):
            configfile = test_path
        if configfile:
            log_msgs['info'].append("%s: Read from configfile:config.yml in installadtion_dir:%s" % (fname, INSTALL_DIR))

    if configfile:
        config, err = parse_config(configfile)

        # MTH: add way to read db specific params from another file:
        if 'DB_NAME' not in config and 'DB_CONFIGFILE' in config:
            _config, _err = parse_config(config['DB_CONFIGFILE'])
            for k,v in _config.items():
                config[k] = v

        if config is None:
            log_msgs['error'].append(err)
            return None, log_msgs

        # MTH: I could allow user to specify the entire db connection via user ENV vars
        #      but this code is used by apps that modify (insert/delete) a db
        #      hence I prefer to force user to explicitly specify db HOST + NAME in config file
        #      so they don't accidentally modify the wrong db
        #for param in ['DB_HOST', 'DB_NAME', 'DB_PORT', 'DB_USER', 'DB_PASSWORD']
        if 'DB_HOST' in config and 'DB_NAME' in config:
            missing_param = False
            for param in ['DB_PORT', 'DB_USER', 'DB_PASSWORD']:
                if param not in config:
                    val = os.getenv(param, None)
                    if val:
                        config[param] = val
                    else:
                        #print("ERROR: param=%s is not specified in configfile or in ENV --> Exitting!" % param)
                        log_msgs['error'].append("param=%s is not specified in configfile or in ENV --> Exitting!" % param)
                        missing_param = True
                        #exit(2)
            if missing_param:
                return None, log_msgs

            config["conn_string"] = "host=%s dbname=%s port=%s user=%s password=%s" % \
                (config['DB_HOST'], config['DB_NAME'], config['DB_PORT'],
                config['DB_USER'], config['DB_PASSWORD'])

            config["sqlalchemy.url"] = "postgresql://{}:{}@{}:{}/{}".format(
                config['DB_USER'], config['DB_PASSWORD'], config['DB_HOST'],
                config['DB_PORT'], config['DB_NAME'])

        return config, log_msgs
    else:
        if requireConfigFile:
            #logger.ERROR("ConfigFile is required but none has been found --> Exit")
            log_msgs['error'].append("ConfigFile is required but none has been found --> Exit")
            log_msgs['info'].append("ConfigFile is searched in the following order:")
            log_msgs['info'].append("1. cmd line: -c/--configfile/--configFile /path/to/myconfig.yml")
            log_msgs['info'].append("2. local dir: ./config.yml")
            log_msgs['info'].append("3. ENV(AQMS_DB_CONFIG) --> points to config file (e.g., >export AQMS_DB_CONFIG=/path/to/config.yml")
            log_msgs['info'].append("4. PKG INSALLATION DIR:%s" % INSTALL_DIR)
            #print()
            #exit(2)

    return None, log_msgs


import argparse
def pick_off_configFile(argv, debug=False):

    parser = argparse.ArgumentParser(add_help=False)
    optional = parser._action_groups.pop()
    #required = parser.add_argument_group("required arguments")
    parser._action_groups.append(optional) #.
    optional.add_argument('--configFile', '--configfile', '-c', type=str, metavar='// --configfile config.yml')
    #optional.add_argument('--configFile', type=str, metavar='--configFile config.yml')
    #optional.add_argument('--configfile', type=str, metavar='--configfile config.yml')
    #optional.add_argument('-c', type=str, metavar='-c config.yml')

    args, unknown = parser.parse_known_args()

    found = False
    index = 0
    if args.configFile:
        if debug:
            print("pick_off_configFile --> args.configFile=%s passed in" % args.configFile)

        return args.configFile

    return None

        #if not os.path.exists(args.configFile):
            #print("ERROR: configfile=[%s] does not exist --> Exitting!" % args.configFile)
            #exit(2)


def parse_config(filename=None):
    """
        Read in config from yaml filename
    """
    configuration = {}
    '''
    if not os.path.exists(filename):
        print("configFile=[%s] does not exist --> Exit" % filename)
        logger.error("configFile=[%s] does not exist --> Exit" % filename)
        exit(2)
    '''

    try:
        with open(filename, 'r') as ymlfile:
            configuration = yaml.load(ymlfile, Loader=yaml.FullLoader)
    except FileNotFoundError as e:
        return None, e
        #return None, "Error trying to read filename:%s: %s" % (filename, e)

    return configuration, None

